@@ -1,5 +1,7 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003 Sergey Poznyakoff
+   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
  
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -34,12 +36,49 @@
 # define DICT_INDEX_SIZE 2048
 #endif
 
+enum dict_symbol_type {
+	dict_symbol_uninitialized,
+	dict_symbol_attribute,
+	dict_symbol_alias
+};
+
+typedef struct dict_symbol DICT_SYMBOL;
+struct dict_symbol {
+        DICT_SYMBOL *next;          /* Link to the next attribute */
+	char *name;
+	enum dict_symbol_type type; /* Type of the entry */
+	union {
+		DICT_ATTR attr;
+		DICT_ATTR *alias;
+	} v;
+};
+
 static Symtab    *dict_attr_tab;
 static DICT_ATTR *dict_attr_index[DICT_INDEX_SIZE];
-static LIST /* of DICT_VALUE */ *dictionary_values;
-static LIST /* of DICT_VENDOR */ *dictionary_vendors;
+static RAD_LIST /* of DICT_VALUE */ *dictionary_values;
+static RAD_LIST /* of DICT_VENDOR */ *dictionary_vendors;
 static int         vendorno;
 
+static DICT_ATTR *
+dict_attr_lookup(char *ident)
+{
+	DICT_SYMBOL *sym = sym_lookup(dict_attr_tab, ident);
+	if (sym) {
+		switch (sym->type) {
+		case dict_symbol_uninitialized:
+			insist_fail("sym_lookup returned uninitialized symbol!");
+			break;
+			
+		case dict_symbol_attribute:
+			return &sym->v.attr;
+
+		case dict_symbol_alias:
+			return sym->v.alias;
+		}
+	}
+	return NULL;
+}
+
 /* ************************************************************************ */
 
 static int
@@ -67,7 +106,7 @@ dict_free()
         if (dict_attr_tab)
                 symtab_clear(dict_attr_tab);
         else
-                dict_attr_tab = symtab_create(sizeof(DICT_ATTR), NULL);
+                dict_attr_tab = symtab_create(sizeof(DICT_SYMBOL), NULL);
         memset(dict_attr_index, 0, sizeof dict_attr_index);
 
 	list_destroy(&dictionary_values, free_value, NULL);
@@ -76,13 +115,13 @@ dict_free()
 }
 
 static int
-nfields(int fc, int minf, int maxf, char *file, int lineno)
+nfields(int fc, int minf, int maxf, LOCUS *loc)
 {
         if (fc < minf) {
-                radlog(L_ERR, "%s:%d: %s", file, lineno, _("too few fields"));
+                radlog_loc(L_ERR, loc, "%s", _("too few fields"));
                 return -1;
         } else if (fc > maxf) {
-                radlog(L_ERR, "%s:%d: %s", file, lineno, _("too many fields"));
+                radlog_loc(L_ERR, loc, "%s", _("too many fields"));
                 return -1;
         }
         return 0;
@@ -91,7 +130,7 @@ nfields(int fc, int minf, int maxf, char *file, int lineno)
 /*
  *      Add vendor to the list.
  */
-int
+static int
 addvendor(char *name, int value)
 {
         DICT_VENDOR *vval;
@@ -119,9 +158,8 @@ struct attr_parser_tab {
 	attr_parser_fp fun;
 };
 static ATTR_PARSER_TAB *attr_parser_tab;
-static attr_parser_fp dict_find_parser(int);
 
-attr_parser_fp
+static attr_parser_fp
 dict_find_parser(int attr)
 {
 	ATTR_PARSER_TAB *ep;
@@ -159,11 +197,11 @@ dict_register_parser(int attr, attr_parser_fp fun)
 #define VENDOR_NAME  fv[1]
 #define VENDOR_VALUE fv[2]
 
-static int _dict_include(int *, int, char **, char *, int);
-static int _dict_attribute(int *, int, char **, char *, int);
-static int _dict_value(int *, int, char **, char *, int);
-static int _dict_vendor(int *, int, char **, char *, int);
-static int parse_dict_entry(int *, int, char **, char *, int);
+static int _dict_include(int *, int, char **, LOCUS *);
+static int _dict_attribute(int *, int, char **, LOCUS *);
+static int _dict_value(int *, int, char **, LOCUS *);
+static int _dict_vendor(int *, int, char **, LOCUS *);
+static int parse_dict_entry(void *, int, char **, LOCUS *);
 static int parse_dict(char *name);
 
 static struct keyword type_kw[] = {
@@ -174,26 +212,24 @@ static struct keyword type_kw[] = {
 };
 
 /*ARGSUSED*/
-int
-_dict_include(int *errcnt, int fc, char **fv, char *file, int lineno)
+static int
+_dict_include(int *errcnt, int fc, char **fv, LOCUS *loc)
 {
-        if (nfields(fc, 2, 2, file, lineno)) 
+        if (nfields(fc, 2, 2, loc)) 
                 return 0;
         parse_dict(fv[1]);
         return 0;
 }
 
-int
-parse_flags(char **ptr, int *flags, char *filename, int line)
+static int
+parse_flags(char **ptr, int *flags, LOCUS *loc)
 {
         int i;
         char *p;
         
         for (p = *ptr+1, i = 0; i < CF_MAX; i++) {
                 if (*p == 0) {
-                        radlog(L_ERR,
-                               _("%s:%d: missing ]"),
-                               filename, line, *p);
+                        radlog_loc(L_ERR, loc, _("missing ]"), *p);
                         return 1;
                 }
                 switch (*p++) {
@@ -208,9 +244,9 @@ parse_flags(char **ptr, int *flags, char *filename, int line)
                         p--;
                         goto stop;
                 default:
-                        radlog(L_ERR,
-                               _("%s:%d: invalid syntax flag %c"),
-                               filename, line, p[-1]);
+                        radlog_loc(L_ERR, loc,
+				   _("invalid syntax flag %c"),
+				   p[-1]);
                         return 1;
                 }
                 switch (*p++) {
@@ -221,9 +257,9 @@ parse_flags(char **ptr, int *flags, char *filename, int line)
                         *flags &= ~AF_RHS(i);
                         break;
                 default:
-                        radlog(L_ERR,
-                               _("%s:%d: invalid syntax flag %c"),
-                               filename, line, p[-1]);
+                        radlog_loc(L_ERR, loc,
+				   _("invalid syntax flag %c"),
+				   p[-1]);
                         return 1;
                 }
         }
@@ -234,19 +270,112 @@ parse_flags(char **ptr, int *flags, char *filename, int line)
         return 0;
 }
 
-int
-_dict_attribute(int *errcnt, int fc, char **fv, char *file, int lineno)
+static int
+parse_attr_properties(LOCUS *loc, char *str, int *flags, int *prop)
 {
+	int errcnt = 0;
+	char *p;
+	
+	for (p = str; *p; p++) {
+		switch (*p) {
+		case 'C':
+		case 'L':
+			*flags |= AF_LHS(CF_USERS)
+				   |AF_LHS(CF_HINTS)
+				   |AF_LHS(CF_HUNTGROUPS);
+			break;
+		case 'R':
+			*flags |= AF_RHS(CF_USERS)
+				   |AF_RHS(CF_HINTS)
+				   |AF_RHS(CF_HUNTGROUPS);
+			break;
+		case '[':
+			if (parse_flags(&p, flags, loc)) {
+				while (*++p);
+				--p;
+				errcnt++;
+			}
+			break;
+		case '=':
+			SET_ADDITIVITY(*prop, AP_ADD_REPLACE);
+			break;
+		case '+':
+			SET_ADDITIVITY(*prop, AP_ADD_APPEND);
+			break;
+		case 'N':
+			SET_ADDITIVITY(*prop, AP_ADD_NONE);
+			break;
+		case 'P':
+			*prop |= AP_PROPAGATE;
+			break;
+		case 'l':
+			*flags &= ~AP_INTERNAL;
+			break;
+		case '1':
+		case '2':
+		case '3':
+		case '4':
+		case '5':
+		case '6':
+		case '7':
+		case '8':
+		case '9':
+			*prop |= AP_USER_FLAG(*p-'0');
+			break;
+		case 'b':
+			*prop |= AP_BINARY_STRING;
+			break;
+		case 'E':
+			*prop |= AP_ENCRYPT_RFC2138;
+			break;
+		case 'T':
+			*prop |= AP_ENCRYPT_RFC2868;
+			break;
+		case 'c':
+			/* Retained for compatibility */
+			break;
+		default:
+			radlog_loc(L_ERR, loc,
+				   _("invalid flag %c"),
+				   *p);
+			errcnt++;
+			break;
+		}
+	}
+	return errcnt;
+}
+
+static void
+set_default_attr_properties(int value, int *flags, int *prop)
+{
+        *flags = AF_DEFAULT_FLAGS;
+        *prop  = AP_DEFAULT_ADD;
+
+	if (VENDOR(value) == 0) {
+		if (value > 255)
+			*flags |= AP_INTERNAL;
+		/* FIXME: A temporary hack until all users update
+		   their dictionaries */
+		else if (value == DA_USER_PASSWORD
+			 || value == DA_USER_PASSWORD)
+			*prop |= AP_ENCRYPT_RFC2138;
+	}
+}
+
+static int
+_dict_attribute(int *errcnt, int fc, char **fv, LOCUS *loc)
+{
+	DICT_SYMBOL *sym;
         DICT_ATTR *attr;
         int type;
         int vendor = 0;
         unsigned value;
         char *p;
-        int flags = AF_DEFAULT_FLAGS;
-        int prop  = AP_DEFAULT_ADD;
 	attr_parser_fp fp = NULL;
+        int flags;
+        int prop;
         
-        if (nfields(fc, 4, 6, file, lineno))
+        if (nfields(fc, 4, 6, loc))
                 return 0;
         /*
          * Validate all entries
@@ -254,9 +383,9 @@ _dict_attribute(int *errcnt, int fc, char **fv, char *file, int lineno)
         
         value = strtol(ATTR_VALUE, &p, 0);
         if (*p) {
-                radlog(L_ERR,
-                       _("%s:%d: value not a number (near %s)"),
-                       file, lineno, p);
+                radlog_loc(L_ERR, loc,
+			   _("value not a number (near %s)"),
+			   p);
                 (*errcnt)++;
                 return 0;
         }
@@ -265,131 +394,160 @@ _dict_attribute(int *errcnt, int fc, char **fv, char *file, int lineno)
 		type = TYPE_STRING;
 		fp = dict_find_parser(value);
 		if (!fp) {
-			radlog(L_WARN,
-		       _("%s:%d: no parser registered for this attribute"),
-			       file, lineno);
+			radlog_loc(L_WARN, loc,
+			      _("no parser registered for this attribute"));
 			return 0;
 		}
 	} else
 		type = xlat_keyword(type_kw, ATTR_TYPE, TYPE_INVALID);
 	
         if (type == TYPE_INVALID) {
-                radlog(L_ERR,
-                       "%s:%d: %s",
-                       file, lineno,
-		       _("invalid type"));
+                radlog_loc(L_ERR, loc,
+			   "%s",
+			   _("invalid type"));
                 (*errcnt)++;
                 return 0;
         }
 
         if (HAS_VENDOR(fc, fv)) {
                 if ((vendor = vendor_name_to_id(ATTR_VENDOR)) == 0) {
-                        radlog(L_ERR,
-                               _("%s:%d: unknown vendor"),
-                               file, lineno);
-                        (*errcnt)++;
+                        radlog_loc(L_ERR, loc, _("unknown vendor"));
+			(*errcnt)++;
                         return 0;
                 }
-        }
+		value |= (vendor << 16);
+        } 
+	set_default_attr_properties(value, &flags, &prop);
 
         if (HAS_FLAGS(fc,fv)) {
-                char *p;
-
-                for (p = ATTR_FLAGS; *p; p++) {
-                        switch (*p) {
-                        case 'C':
-                        case 'L':
-                                flags |= AF_LHS(CF_USERS)
-                                        |AF_LHS(CF_HINTS)
-                                        |AF_LHS(CF_HUNTGROUPS);
-                                break;
-                        case 'R':
-                                flags |= AF_RHS(CF_USERS)
-                                        |AF_RHS(CF_HINTS)
-                                        |AF_RHS(CF_HUNTGROUPS);
-                                break;
-                        case '[':
-                                if (parse_flags(&p, &flags, file, lineno)) {
-                                        while (*++p);
-                                        --p;
-                                        ++(*errcnt);
-                                }
-                                break;
-                        case '=':
-                                SET_ADDITIVITY(prop, AP_ADD_REPLACE);
-                                break;
-                        case '+':
-                                SET_ADDITIVITY(prop, AP_ADD_APPEND);
-                                break;
-                        case 'N':
-                                SET_ADDITIVITY(prop, AP_ADD_NONE);
-                                break;
-                        case 'P':
-                                prop |= AP_PROPAGATE;
-                                break;
-			case 'c':
-				prop |= AP_REQ_CMP;
-				break;
-			case '1':
-			case '2':
-			case '3':
-			case '4':
-			case '5':
-			case '6':
-			case '7':
-			case '8':
-			case '9':
-				prop |= AP_USER_FLAG(*p-'0');
-				break;
-			case 'b':
-				prop |= AP_BINARY_STRING;
-                                break;
-                        default:
-                                radlog(L_ERR,
-                                       _("%s:%d: invalid flag %c"),
-                                       file, lineno, *p);
-                                (*errcnt)++;
-                                return 0;
-                        }
-                }
+		int rc = parse_attr_properties(loc, ATTR_FLAGS, &flags, &prop);
+		if (rc) {
+			++*errcnt;
+			return 0;
+		}
         }
 
-        attr = sym_lookup_or_install(dict_attr_tab, ATTR_NAME, 1);
-                        
+	sym = sym_lookup_or_install(dict_attr_tab, ATTR_NAME, 1);
+	switch (sym->type) {
+	case dict_symbol_uninitialized:
+		sym->type = dict_symbol_attribute;
+		attr = &sym->v.attr;
+		break;
+
+	case dict_symbol_attribute:
+		radlog_loc(L_WARN, loc,
+			   _("Redefining attribute %s"),
+			   ATTR_NAME);
+		attr = &sym->v.attr;
+		break;
+		
+	case dict_symbol_alias:
+		radlog_loc(L_WARN, loc,
+			   _("Redefining alias %s"),
+			   ATTR_NAME);
+		attr = sym->v.alias;
+	}
+
+	attr->name = sym->name;
         attr->value = value;
         attr->type = type;
-        attr->prop = flags|prop;
+        attr->prop = flags | prop;
 	attr->parser = fp;
-        if (vendor)
-                attr->value |= (vendor << 16);
-        if (attr->value < DICT_INDEX_SIZE) 
+        if (attr->value >= 0 && attr->value < DICT_INDEX_SIZE) 
                 dict_attr_index[attr->value] = attr;
         
         return 0;
 }
 
-int
-_dict_value(int *errcnt, int fc, char **fv, char *file, int lineno)
+/* Syntax:
+   ALIAS oldname newname */
+static int
+_dict_alias(int *errcnt, int fc, char **fv, LOCUS *loc)
+{
+	DICT_SYMBOL *sym;
+	DICT_ATTR *attr;
+	
+	if (nfields(fc, 3, 3, loc))
+                return 0;
+
+        attr = dict_attr_lookup(fv[1]);
+	if (!attr) {
+		radlog_loc(L_ERR, loc,
+			   _("Attribute %s is not defined"),
+			   fv[1]);
+		return 0;
+	}
+		
+	sym = sym_lookup_or_install(dict_attr_tab, fv[2], 1);
+	if (sym->type != dict_symbol_uninitialized) {
+		radlog_loc(L_ERR, loc,
+			   _("Symbol %s already declared"),
+			   fv[2]);
+		/*FIXME: location of the previous declaration*/
+		/* Not a fatal error: do not increase error count */
+		return 0;
+	}
+	sym->type = dict_symbol_alias;
+	sym->v.alias = attr;
+	return 0;
+}
+
+/* Syntax:
+   
+   PROPERTY Attribute Flags */
+static int
+_dict_property(int *errcnt, int fc, char **fv, LOCUS *loc)
+{
+	DICT_ATTR *attr;
+	int flags;
+	int prop;
+	
+	if (nfields(fc, 3, 3, loc))
+                return 0;
+
+        attr = dict_attr_lookup(fv[1]);
+	if (!attr) {
+		radlog_loc(L_ERR, loc,
+			   _("Attribute %s is not defined"),
+			   fv[1]);
+		return 0;
+	}
+	set_default_attr_properties(attr->value, &flags, &prop);
+	if (parse_attr_properties(loc, fv[2], &flags, &prop))
+		++*errcnt;
+        attr->prop = flags | prop;
+	return 0;
+}
+
+static int
+_dict_value(int *errcnt, int fc, char **fv, LOCUS *loc)
 {
         DICT_VALUE *dval;
         DICT_ATTR *attr;
         char *p;
         int value;
         
-        if (nfields(fc, 4, 4, file, lineno))
+        if (nfields(fc, 4, 4, loc))
                 return 0;
 
         value = strtol(VALUE_NUM, &p, 0);
         if (*p) {
-                radlog(L_ERR,
-                       _("%s:%d: value not a number (near %s)"),
-                       file, lineno, p);
+                radlog_loc(L_ERR, loc,
+			   _("value not a number (near %s)"),
+			   p);
                 (*errcnt)++;
                 return 0;
         }
 
-        attr = sym_lookup_or_install(dict_attr_tab, VALUE_ATTR, 1);
-        
+        attr = dict_attr_lookup(VALUE_ATTR);
+	if (!attr) {
+		radlog_loc(L_ERR, loc,
+			   _("Attribute %s is not defined"),
+			   VALUE_ATTR);
+                (*errcnt)++;
+		return 0;
+	}
+	
         /* Create a new VALUE entry for the list */
         dval = emalloc(sizeof(DICT_VALUE));
                         
@@ -405,20 +563,20 @@ _dict_value(int *errcnt, int fc, char **fv, char *file, int lineno)
         return 0;
 }
 
-int
-_dict_vendor(int *errcnt, int fc, char **fv, char *file, int lineno)
+static int
+_dict_vendor(int *errcnt, int fc, char **fv, LOCUS *loc)
 {
         int value;
         char *p;
 
-        if (nfields(fc, 3, 3, file, lineno))
+        if (nfields(fc, 3, 3, loc))
                 return 0;
 
         value = strtol(VENDOR_VALUE, &p, 0);
         if (*p) {
-                radlog(L_ERR,
-                       _("%s:%d: value not a number (near %s)"),
-                       file, lineno, p);
+                radlog_loc(L_ERR, loc,
+			   _("value not a number (near %s)"),
+			   p);
                 (*errcnt)++;
                 return 0;
         }
@@ -434,39 +592,47 @@ _dict_vendor(int *errcnt, int fc, char **fv, char *file, int lineno)
 enum {
         KW_INCLUDE,
         KW_ATTRIBUTE,
+	KW_ALIAS,
         KW_VALUE,
-        KW_VENDOR
+        KW_VENDOR,
+	KW_PROPERTY
 };
 
 static struct keyword dict_kw[] = {
         { "$INCLUDE", KW_INCLUDE },
         { "ATTRIBUTE", KW_ATTRIBUTE },
+	{ "ALIAS", KW_ALIAS },
         { "VALUE", KW_VALUE },
         { "VENDOR", KW_VENDOR },
+	{ "PROPERTY", KW_PROPERTY },
         { NULL, 0 }
 };
 
 int
-parse_dict_entry(int *errcnt, int fc, char **fv, char *file, int lineno)
+parse_dict_entry(void *closure, int fc, char **fv, LOCUS *loc)
 {
+	int *errcnt = closure;
         switch (xlat_keyword(dict_kw, KEYWORD, -1)) {
         case KW_INCLUDE:
-                _dict_include(errcnt, fc, fv, file, lineno);
+                _dict_include(errcnt, fc, fv, loc);
                 break;
         case KW_ATTRIBUTE:
-                _dict_attribute(errcnt, fc, fv, file, lineno);
+                _dict_attribute(errcnt, fc, fv, loc);
                 break;
+	case KW_ALIAS:
+		_dict_alias(errcnt, fc, fv, loc);
+		break;
         case KW_VALUE:
-                _dict_value(errcnt, fc, fv, file, lineno);
+                _dict_value(errcnt, fc, fv, loc);
                 break;
         case KW_VENDOR:
-                _dict_vendor(errcnt, fc, fv, file, lineno);
+                _dict_vendor(errcnt, fc, fv, loc);
                 break;
+	case KW_PROPERTY:
+		_dict_property(errcnt, fc, fv, loc);
+		break;
         default:
-                radlog(L_ERR,
-                       "%s:%d: %s",
-                       file, lineno,
-		       _("unknown keyword"));
+                radlog_loc(L_ERR, loc, "%s", _("unknown keyword"));
                 break;
         }
         return 0;
@@ -478,8 +644,11 @@ parse_dict(char *name)
         char *path;
         int   rc;
         int   errcnt = 0;
-        
-        path = mkfilename(radius_dir, name);
+
+	if (name[0] == '/')
+		path = estrdup(name);
+	else
+		path = mkfilename(radius_dir, name);
         rc = read_raddb_file(path, 1, parse_dict_entry, &errcnt);
         if (errcnt)
                 radlog(L_NOTICE,
@@ -516,9 +685,24 @@ struct attr_value {
 };
 
 int
-attrval_cmp(struct attr_value *av, DICT_ATTR *attr)
+attrval_cmp(struct attr_value *av, DICT_SYMBOL *sym)
 {
-        if (attr->value == av->value) {
+	DICT_ATTR *attr;
+	
+	switch (sym->type) {
+	case dict_symbol_attribute:
+		attr = &sym->v.attr;
+		break;
+		
+	case dict_symbol_alias:
+		attr = sym->v.alias;
+		break;
+
+	default:
+		return 0;
+	}
+
+	if (attr->value == av->value) {
                 av->da = attr;
                 return 1;
         }
@@ -529,7 +713,7 @@ DICT_ATTR *
 attr_number_to_dict(int attribute)
 {
         struct attr_value av;
-        if (attribute < DICT_INDEX_SIZE)
+        if (attribute >= 0 && attribute < DICT_INDEX_SIZE)
                 return dict_attr_index[attribute];
         av.value = attribute;
         av.da = NULL;
@@ -544,7 +728,7 @@ attr_number_to_dict(int attribute)
 DICT_ATTR *
 attr_name_to_dict(char *attrname)
 {
-        return sym_lookup(dict_attr_tab, attrname);
+        return dict_attr_lookup(attrname);
 }
 
 /*
