@@ -1,5 +1,7 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2002,2003 Sergey Poznyakoff
+   Copyright (C) 2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
   
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -80,7 +82,7 @@ rad_pdu_finish(void **ptr, struct radius_pdu *pdu,
         size_t len = sizeof(AUTH_HDR) + pdu->size;
         u_char digest[AUTH_DIGEST_LEN];
         
-	if (code != RT_AUTHENTICATION_REQUEST) {
+	if (code != RT_ACCESS_REQUEST && code != RT_STATUS_SERVER) {
                 secretlen = strlen(secret);
                 obstack_grow(&pdu->st, secret, secretlen);
         }
@@ -96,7 +98,8 @@ rad_pdu_finish(void **ptr, struct radius_pdu *pdu,
 	/* Seal the message properly. Note that the secret has already been
 	   appended to the pdu wherever necessary */
 	switch (code) {
-	case RT_AUTHENTICATION_REQUEST:
+	case RT_ACCESS_REQUEST:
+	case RT_STATUS_SERVER:
 		memcpy(hdr->vector, vector, AUTH_VECTOR_LEN);
 		break;
 		
@@ -110,8 +113,8 @@ rad_pdu_finish(void **ptr, struct radius_pdu *pdu,
                 memset((char*)hdr + len, 0, secretlen);
 		break;
 		
-	case RT_AUTHENTICATION_ACK:
-	case RT_AUTHENTICATION_REJECT:
+	case RT_ACCESS_ACCEPT:
+	case RT_ACCESS_REJECT:
 	case RT_ACCOUNTING_RESPONSE:
 	case RT_ACCESS_CHALLENGE:
 		memcpy(hdr->vector, vector, AUTH_VECTOR_LEN);
@@ -220,23 +223,30 @@ rad_create_pdu(void **rptr, int code, int id, u_char *vector,
                 if (debug_on(10)) {
                         char *save;
                         radlog(L_DEBUG,
-                               "send: %s", format_pair(pair, &save));
+                               "send: %s", format_pair(pair, 1, &save));
                         free(save);
                 }
 
                 rad_attr_init(&attr);
                 if ((vendorcode = VENDOR(pair->attribute)) > 0
                     && (vendorpec  = vendor_id_to_pec(vendorcode)) > 0) {
-                        u_char c;
-                        
                         attr.attrno = DA_VENDOR_SPECIFIC;
                         lval = htonl(vendorpec);
                         rad_attr_write(&attr, &lval, 4);
-                        c = pair->attribute & 0xff;
-                        rad_attr_write(&attr, &c, 1); 
-                        rad_attr_write(&attr, &lval, 1); /* Reserve a byte */
-                        attrlen = rad_encode_pair(&attr, pair);
-                        attr.data[5] = 2+attrlen; /* Fill in the length */
+			if (vendorpec == 429) {
+				/* Hack for non-compliant USR VSA */
+				UINT4 atval = htonl(pair->attribute & 0xffff);
+				rad_attr_write(&attr, &atval, 4);
+				attrlen = rad_encode_pair(&attr, pair);
+			} else {
+				u_char c = pair->attribute & 0xff;
+				rad_attr_write(&attr, &c, 1);
+				/* Reserve a length byte */
+				rad_attr_write(&attr, &lval, 1); 
+				attrlen = rad_encode_pair(&attr, pair);
+				/* Fill in the length */
+				attr.data[5] = 2+attrlen; 
+			}
                 } else if (pair->attribute > 0xff)
                         continue;
 		else {
@@ -326,7 +336,7 @@ rad_decode_pair(UINT4 attrno, char *ptr, size_t attrlen)
                 if (debug_on(10)) {
                         char *save;
                         radlog(L_DEBUG, "recv: %s",
-                               format_pair(pair, &save));
+                               format_pair(pair, 1, &save));
                         free(save);
                 }
 
@@ -341,7 +351,7 @@ rad_decode_pair(UINT4 attrno, char *ptr, size_t attrlen)
                         char *save;
                         radlog(L_DEBUG, 
                                "recv: %s", 
-                               format_pair(pair, &save));
+                               format_pair(pair, 1, &save));
                         free(save);
                 }
                 break;
@@ -452,9 +462,11 @@ rad_decode_pdu(UINT4 host, u_short udp_port, u_char *buffer, size_t length)
 					ptr += 4;
 					attrlen -= 4;
 					len = attrlen;
+					attrlen = 0;
 				} else {
 					attrno = *ptr++ | (vendorcode << 16);
 					len = *ptr++ - 2;
+					attrlen -= len + 2;
 				}
 				
                                 pair = rad_decode_pair(attrno, ptr, len);
@@ -466,7 +478,6 @@ rad_decode_pdu(UINT4 host, u_short udp_port, u_char *buffer, size_t length)
                                 	prev = pair;
 				} 
                                 ptr += len;
-                                attrlen -= len + 2;
                         }
                 } else {
                         pair = rad_decode_pair(attrno, ptr, attrlen);
