@@ -1,5 +1,7 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003 Sergey Poznyakoff
+   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
  
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -371,23 +373,28 @@ rad_acct_system(RADIUS_REQ *radreq, int dowtmp)
         return ret;
 }
 
-int
+static int
 check_acct_dir()
 {
-        struct stat     st;
+        struct stat st;
 
         if (stat(radacct_dir, &st) == 0) {
-                if (S_ISDIR(st.st_mode))
+                if (S_ISDIR(st.st_mode)) {
+			if (access(radacct_dir, W_OK)) {
+				radlog(L_ERR,
+				       _("%s: directory not writable"),
+				       radacct_dir);
+				return 1;
+			}
                         return 0;
-                else {
-                        radlog(L_ERR,
-                                _("%s: not a directory"),
-                                radacct_dir);
+		} else {
+                        radlog(L_ERR, _("%s: not a directory"),
+			       radacct_dir);
                         return 1;
                 }
         }
         if (mkdir(radacct_dir, 0755)) {
-                radlog(L_CRIT|L_PERROR,
+                radlog(L_ERR|L_PERROR,
                         _("can't create accounting directory `%s'"),
                         radacct_dir);
                 return 1;
@@ -395,23 +402,43 @@ check_acct_dir()
         return 0;
 }
 
+static int acct_dir_status;
+
+static void
+acct_after_config_hook(void *arg ARG_UNUSED, void *data ARG_UNUSED)
+{
+	if (auth_detail || acct_detail) {
+		acct_dir_status = check_acct_dir();
+		if (acct_dir_status) {
+			radlog(L_NOTICE,
+			       _("Detailed accounting is disabled"));
+			auth_detail = acct_detail = 0;
+		}
+	}
+}
+
+void
+acct_init()
+{
+	radiusd_set_postconfig_hook(acct_after_config_hook, NULL, 0);
+}
+
 int
 write_detail(RADIUS_REQ *radreq, int authtype, char *f)
 {
-        FILE            *outfd;
-        char            nasname[MAX_LONGNAME];
-        char            *dir, *path;
-        char            *save;
-        VALUE_PAIR      *pair;
-        UINT4           nas;
-        NAS             *cl;
-        time_t          curtime;
-        int             ret = 0;
-        struct stat     st;
+        FILE *outfd;
+        char nasname[MAX_LONGNAME];
+        char *dir, *path;
+        char *save;
+        VALUE_PAIR *pair;
+        UINT4 nas;
+        NAS *cl;
+        time_t curtime;
+        int ret = 0;
 
-        /* A superfluous precaution, maybe: */
-        if (stat(radacct_dir, &st) < 0)
-                return 0;
+	if (acct_dir_status)
+		return;
+	
         curtime = time(0);
         
         /* Find out the name of this terminal server. We try to find
@@ -466,14 +493,15 @@ write_detail(RADIUS_REQ *radreq, int authtype, char *f)
                                 pair = avl_find(radreq->request, DA_USER_NAME);
                         if (pair) {
                                 fprintf(outfd, "\t%s\n", 
-                                        format_pair(pair, &save));
+                                        format_pair(pair, 0, &save));
                                 free(save);
                         }
                 }
 
                 /* Write each attribute/value to the log file */
-                pair = radreq->request;
-                while (pair != (VALUE_PAIR *)NULL) {
+                for (pair = radreq->request; pair; pair = pair->next) {
+			if (pair->prop & AP_INTERNAL)
+				continue;
                         switch (pair->attribute) {
                         case DA_USER_PASSWORD:
                                 break;
@@ -483,28 +511,29 @@ write_detail(RADIUS_REQ *radreq, int authtype, char *f)
                                         break;
                         default:
                                 fprintf(outfd, "\t%s\n", 
-                                        format_pair(pair, &save));
+                                        format_pair(pair, 0, &save));
                                 free(save);
                         } 
-                        pair = pair->next;
                 }
 
                 /* Add non-protocol attibutes. */
                 fprintf(outfd, "\tTimestamp = %ld\n", curtime);
                 switch (authtype) {
-                    case REQ_AUTH_OK:
-                        fprintf( outfd, "\tRequest-Authenticator = Verified\n");
+		case REQ_AUTH_OK:
+			fprintf(outfd, "\tRequest-Authenticator = Verified\n");
                         break;
-                    case REQ_AUTH_ZERO:
-                        fprintf( outfd, "\tRequest-Authenticator = None\n");
+		case REQ_AUTH_ZERO:
+                        fprintf(outfd, "\tRequest-Authenticator = None\n");
                         break;
-                    case REQ_AUTH_BAD:
-                        fprintf( outfd, "\tRequest-Authenticator = Unverified\n");
+		case REQ_AUTH_BAD:
+                        fprintf(outfd, "\tRequest-Authenticator = Unverified\n");
                         break;
-                    default:
+		default:
+			fprintf(outfd, "\tRequest-Authenticator = %d\n",
+				authtype);
                         break;
                 }
-                fprintf( outfd, "\n");
+                fprintf(outfd, "\n");
                 fclose(outfd);
                 ret = 0;
         }
@@ -558,7 +587,7 @@ rad_accounting(RADIUS_REQ *radreq, int activefd, int verified)
 {
 	log_open(L_ACCT);
 
-        huntgroup_access(radreq);
+        huntgroup_access(radreq, NULL);
 
 #if defined(RT_ASCEND_EVENT_REQUEST) && defined(RT_ASCEND_EVENT_RESPONSE)
         /* Special handling for Ascend-Event-Request */
@@ -623,7 +652,7 @@ check_ts(struct radutmp *ut)
         /* Find NAS type. */
         if ((nas = nas_lookup_ip(ntohl(ut->nas_address))) == NULL) {
                 radlog(L_NOTICE, _("check_ts(): unknown NAS"));
-                return 0; 
+                return -1; 
         }
 
         /* Handle two special types */
@@ -641,8 +670,9 @@ rad_check_realm(REALM *realm)
         int count;
         struct radutmp *up;
         radut_file_t file;
-        
-        if (!realm || realm->maxlogins == 0)
+        size_t maxlogins;
+		
+        if (!realm || (maxlogins = realm_get_quota(realm)) == 0)
                 return 0;
 
         if ((file = rut_setent(radutmp_path, 0)) == NULL)
@@ -655,7 +685,7 @@ rad_check_realm(REALM *realm)
 		    && realm_verify_ip(realm, up->realm_address))
                         count++;
 
-        if (count < realm->maxlogins) {
+        if (count < maxlogins) {
                 rut_endent(file);
                 return 0;
         }
@@ -681,7 +711,7 @@ rad_check_realm(REALM *realm)
                 }
         }
         rut_endent(file);
-        return !(count < realm->maxlogins);
+        return !(count < maxlogins);
 }
 
 /* See if a user is already logged in.
