@@ -1,6 +1,8 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003 Sergey Poznyakoff
-  
+   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
+
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
@@ -53,16 +55,45 @@
 #include <raddbm.h>
 #include <obstack1.h>
 
+typedef struct locus_name {
+	struct locus_name *next;
+	char *filename;
+	size_t refcnt;
+} Locus_symbol;
+
+static Symtab *locus_tab;
+
+void
+locus_dup(LOCUS *dst, LOCUS *src)
+{
+	Locus_symbol *sp;
+	
+	if (!locus_tab)
+                locus_tab = symtab_create(sizeof(Locus_symbol), NULL);
+	sp = (Locus_symbol*) sym_lookup_or_install(locus_tab, src->file, 1);
+	sp->refcnt++;
+	dst->file = sp->filename;
+	dst->line = src->line;
+}
+
+void
+locus_free(LOCUS *loc)
+{
+	Locus_symbol *sp = (Locus_symbol*) sym_lookup(locus_tab, loc->file);
+	if (--sp->refcnt == 0)
+		symtab_delete(locus_tab, (Symbol*) sp);
+}
+
 /*
  * Symbol tables and lists
  */
 Symtab          *user_tab;     /* raddb/users  */
 Symtab          *deny_tab;     /* raddb/access.deny */
 
-MATCHING_RULE   *huntgroups;   /* raddb/huntgroups */ 
-MATCHING_RULE   *hints;        /* raddb/hints */
-LIST /* of CLIENT */ *clients; /* raddb/clients */
-LIST /* of RADCK_TYPE */ *radck_type;   /* raddb/nastypes */
+static RAD_LIST /* of MATCHING_RULE */  *huntgroups;   /* raddb/huntgroups */ 
+static RAD_LIST /* of MATCHING_RULE */  *hints;        /* raddb/hints */
+static RAD_LIST /* of CLIENT */ *clients; /* raddb/clients */
+static RAD_LIST /* of RADCK_TYPE */ *radck_type;   /* raddb/nastypes */
 
 static struct keyword op_tab[] = {
         { "=", OPERATOR_EQUAL },
@@ -91,7 +122,7 @@ static int user_find_sym(char *name, RADIUS_REQ *req,
 int user_find_db(char *name, RADIUS_REQ *req,
                         VALUE_PAIR **check_pairs, VALUE_PAIR **reply_pairs);
 #endif
-static MATCHING_RULE *file_read(int cf_file, char *name);
+static RAD_LIST *file_read(int cf_file, char *name);
 
 int
 comp_op(int op, int result)
@@ -139,9 +170,10 @@ comp_op(int op, int result)
  * parser
  */
 int
-add_user_entry(Symtab *symtab, char *filename, int line,
+add_user_entry(void *closure, LOCUS *loc,
 	       char *name, VALUE_PAIR *check, VALUE_PAIR *reply)
 {
+	Symtab *symtab = closure;
         User_symbol *sym;
 
         /* Special handling for DEFAULT names: strip any trailing
@@ -153,12 +185,11 @@ add_user_entry(Symtab *symtab, char *filename, int line,
                 name = "BEGIN";
 
         if ((check == NULL && reply == NULL)
-            || fix_check_pairs(CF_USERS, filename, line, name, &check)
-            || fix_reply_pairs(CF_USERS, filename, line, name, &reply)) {
-                radlog(L_ERR,
-                       _("%s:%d: discarding user `%s'"),
-                       filename,
-                       line, name);
+            || fix_check_pairs(CF_USERS, loc, name, &check)
+            || fix_reply_pairs(CF_USERS, loc, name, &reply)) {
+                radlog_loc(L_ERR, loc,
+			   _("discarding user `%s'"),
+			   name);
                 avl_free(check);
                 avl_free(reply);
                 return 0;
@@ -168,52 +199,46 @@ add_user_entry(Symtab *symtab, char *filename, int line,
         
         sym->check = check;
         sym->reply = reply;
-        sym->lineno = line;
+        locus_dup(&sym->loc, loc);
         return 0;
 }
 
 static int
 free_user_entry(User_symbol *sym)
 {
+	locus_free(&sym->loc);
         avl_free(sym->check);
         avl_free(sym->reply);
         return 0;
 }
 
-struct temp_list {
+struct temp_data {
         int cf_file;
-        MATCHING_RULE *head;
-        MATCHING_RULE *tail;
+	RAD_LIST *list;
 };
 
 int
-add_pairlist(struct temp_list *closure, char *filename, int line,
+add_pairlist(void *closure, LOCUS *loc,
 	     char *name, VALUE_PAIR *lhs, VALUE_PAIR *rhs)
 {
-        MATCHING_RULE *pl;
+	struct temp_data *data = closure;
+        MATCHING_RULE *rule;
         
         if ((lhs == NULL && rhs == NULL)
-            || fix_check_pairs(closure->cf_file, filename, line, name, &lhs)
-            || fix_reply_pairs(closure->cf_file, filename, line, name, &rhs)) {
-                radlog(L_ERR,
-                       _("%s:%d: discarding entry `%s'"),
-                       filename,
-                       line, name);
+            || fix_check_pairs(data->cf_file, loc, name, &lhs)
+            || fix_reply_pairs(data->cf_file, loc, name, &rhs)) {
+                radlog_loc(L_ERR, loc, _("discarding entry `%s'"), name);
                 avl_free(lhs);
                 avl_free(rhs);
                 return 0;
         }
 
-        pl = emalloc(sizeof(MATCHING_RULE));
-        pl->name = estrdup(name);
-        pl->lhs = lhs;
-        pl->rhs = rhs;
-        pl->lineno = line;
-        if (closure->tail)
-                closure->tail->next = pl;
-        else
-                closure->head = pl;
-        closure->tail = pl;
+        rule = emalloc(sizeof(MATCHING_RULE));
+        rule->name = estrdup(name);
+        rule->lhs = lhs;
+        rule->rhs = rhs;
+        locus_dup(&rule->loc, loc);
+	list_append(data->list, rule);
         return  0;
 }
 
@@ -225,15 +250,15 @@ read_users(char *name)
         return parse_file(name, user_tab, add_user_entry);
 }
 
-MATCHING_RULE *
+static RAD_LIST *
 file_read(int cf_file, char *name)
 {
-        struct temp_list tmp;
+        struct temp_data tmp;
 
         tmp.cf_file = cf_file;
-        tmp.head = tmp.tail = NULL;
+        tmp.list = list_create();
         parse_file(name, &tmp, add_pairlist);
-        return tmp.head;
+        return tmp.list;
 }
 
 enum lookup_state {
@@ -298,7 +323,7 @@ static int match_user(User_symbol *sym, RADIUS_REQ *req,
  */
 int
 user_find_sym(char *name, RADIUS_REQ *req,
-	      VALUE_PAIR **check_pairs, VALUE_PAIR **reply_pairs)
+              VALUE_PAIR **check_pairs, VALUE_PAIR **reply_pairs)
 {
         int found = 0;
         User_symbol *sym;
@@ -320,7 +345,7 @@ user_find_sym(char *name, RADIUS_REQ *req,
 
 int
 match_user(User_symbol *sym, RADIUS_REQ *req,
-	   VALUE_PAIR **check_pairs, VALUE_PAIR **reply_pairs)
+           VALUE_PAIR **check_pairs, VALUE_PAIR **reply_pairs)
 {
         VALUE_PAIR *p;
         VALUE_PAIR *check_tmp;
@@ -341,6 +366,7 @@ match_user(User_symbol *sym, RADIUS_REQ *req,
                         continue;
                 }
 
+		radius_req_register_locus(req, &sym->loc);
                 found = 1;
 
                 for (p = avl_find(sym->check, DA_MATCH_PROFILE);
@@ -348,7 +374,8 @@ match_user(User_symbol *sym, RADIUS_REQ *req,
                      p = avl_find(p->next, DA_MATCH_PROFILE)) {
                         debug(1, ("submatch: %s", p->avp_strvalue));
 
-                        found = match_user(sym_lookup(user_tab, p->avp_strvalue),
+                        found = match_user(sym_lookup(user_tab,
+						      p->avp_strvalue),
                                            req, check_pairs, reply_pairs);
                 }                       
 
@@ -376,7 +403,8 @@ match_user(User_symbol *sym, RADIUS_REQ *req,
                 }
                 if (!fallthrough(sym->reply))
                         break;
-                debug(1, ("fall through near line %d", sym->lineno));
+                debug(1, ("fall through near %s:%lu",
+			  sym->loc.file, (unsigned long) sym->loc.line));
         } while (sym = sym_next((Symbol*)sym));
 
         return found;
@@ -390,7 +418,7 @@ match_user(User_symbol *sym, RADIUS_REQ *req,
  */
 int
 user_find(char *name, RADIUS_REQ *req,
-	  VALUE_PAIR **check_pairs, VALUE_PAIR **reply_pairs)
+          VALUE_PAIR **check_pairs, VALUE_PAIR **reply_pairs)
 {
         int found = 0;
 
@@ -453,12 +481,12 @@ nextkn(char **sptr, char *token, int toksize)
         /* skip whitespace */
         while (**sptr && isws(**sptr))
                 ++(*sptr);
-        if (!*sptr)
-                return 0;
+        if (!**sptr)
+                return -1;
         start = token;
         if (**sptr == '"') {
                 (*sptr)++;
-                while (toksize && **sptr && **sptr != '"') {
+                while (toksize && **sptr) {
                         if (**sptr == '\\' && (*sptr)[1]) {
                                 switch (*++*sptr) {
                                 default:
@@ -487,6 +515,9 @@ nextkn(char **sptr, char *token, int toksize)
                                 }
                                 ++*sptr;
                                 toksize--;
+                        } else if (**sptr == '"') {
+                                ++*sptr;
+                                break;
                         } else 
                                 *token++ = *(*sptr)++;
                 }
@@ -517,9 +548,13 @@ userparse(char *buffer, VALUE_PAIR **first_pair, char **errmsg)
         int             op;
         static char errbuf[512];
         char token[256];
+	LOCUS loc;
 
+	loc.file = "<stdin>"; /* FIXME */
+	loc.line = 0;
+	
         state = PS_LHS;
-        while (nextkn(&buffer, token, sizeof(token))) {
+        while (nextkn(&buffer, token, sizeof(token)) >= 0) {
                 switch (state) {
                 case PS_LHS:
                         if (token[0] == '\n' || token[0] == '#')
@@ -547,14 +582,13 @@ userparse(char *buffer, VALUE_PAIR **first_pair, char **errmsg)
                         break;
                         
                 case PS_RHS:
-			pair = install_pair("<stdin>", 0, attr->name, 
-                                             op, token);
-			if (!pair) {
-				snprintf(errbuf, sizeof(errbuf),
-				         _("install_pair failed on %s"),
-					 attr->name);
-				return -1;
-			}
+                        pair = install_pair(&loc, attr->name, op, token);
+                        if (!pair) {
+                                snprintf(errbuf, sizeof(errbuf),
+                                         _("install_pair failed on %s"),
+                                         attr->name);
+                                return -1;
+                        }
                         avl_merge(first_pair, &pair);
                         state = PS_END;
                         break;
@@ -578,10 +612,41 @@ userparse(char *buffer, VALUE_PAIR **first_pair, char **errmsg)
  * raddb/hints
  */
 
-/*
- *      Add hints to the info sent by the terminal server
- *      based on the pattern of the username.
- */
+/* Provide a support for backward-compatible attributes Replace-User-Name
+   and Rewrite-Function */
+static void
+hints_eval_compat(RADIUS_REQ *req, VALUE_PAIR *name_pair, MATCHING_RULE *rule)
+{
+        VALUE_PAIR      *tmp;
+
+	/* Let's see if we need to further modify the username */
+	if ((tmp = avl_find(rule->rhs, DA_REPLACE_USER_NAME))
+	    || (tmp = avl_find(rule->lhs, DA_REPLACE_USER_NAME))) {
+		char *ptr;
+		struct obstack hints_stk;
+ 
+		obstack_init(&hints_stk);
+		ptr = radius_xlate(&hints_stk, tmp->avp_strvalue,
+				   req, NULL);
+		if (ptr) 
+			string_replace(&name_pair->avp_strvalue, ptr);
+		obstack_free(&hints_stk, NULL);
+	}
+                
+	/* Is the rewrite function specified? */
+	if ((tmp = avl_find(rule->rhs, DA_REWRITE_FUNCTION))
+	    || (tmp = avl_find(rule->lhs, DA_REWRITE_FUNCTION))) {
+		if (rewrite_eval(tmp->avp_strvalue, req, NULL, NULL)) {
+			radlog_loc(L_ERR, &rule->loc, "%s(): %s",
+				   tmp->avp_strvalue,
+				   _("not defined"));
+		}
+	}
+}
+
+/* Add hints to the info sent by the terminal server
+   based on the pattern of the username and the contents of the incoming
+   request */
 int
 hints_setup(RADIUS_REQ *req)
 {
@@ -590,21 +655,25 @@ hints_setup(RADIUS_REQ *req)
         VALUE_PAIR      *name_pair;
         VALUE_PAIR      *orig_name_pair;
         VALUE_PAIR      *tmp;
-        MATCHING_RULE   *i;
+        MATCHING_RULE   *rule;
         int             matched = 0;
-
-	/* Add Proxy-Replied pair if necessary */
-        switch (req->code) {
-	case RT_AUTHENTICATION_ACK:
-	case RT_AUTHENTICATION_REJECT:
-	case RT_ACCOUNTING_RESPONSE:
-	case RT_ACCESS_CHALLENGE:
-		tmp = avp_create_integer(DA_PROXY_REPLIED, 1);
-		avl_merge(&request_pairs, &tmp);
-		avp_free(tmp);
-		break;
-	}
+	ITERATOR *itr;
 	
+        /* Add Proxy-Replied pair if necessary */
+        switch (req->code) {
+        case RT_ACCESS_ACCEPT:
+        case RT_ACCESS_REJECT:
+        case RT_ACCOUNTING_RESPONSE:
+        case RT_ACCESS_CHALLENGE:
+                tmp = avp_create_integer(DA_PROXY_REPLIED, 1);
+                avl_merge(&request_pairs, &tmp);
+                avp_free(tmp);
+                break;
+
+        case RT_STATUS_SERVER:
+                return 0;
+        }
+        
         if (hints == NULL)
                 return 0;
 
@@ -612,12 +681,12 @@ hints_setup(RADIUS_REQ *req)
          *      Check for valid input, zero length names not permitted 
          */
         if ((name_pair = avl_find(request_pairs, DA_USER_NAME)) == NULL) {
-		name_pair = avp_create_string(DA_USER_NAME, "");
+                name_pair = avp_create_string(DA_USER_NAME, "");
                 orig_name_pair = NULL;
         } else {
                 orig_name_pair = avp_create_string(DA_ORIG_USER_NAME,
-						   name_pair->avp_strvalue);
-	}
+                                                   name_pair->avp_strvalue);
+        }
 
         debug(1, ("called for `%s'", name_pair->avp_strvalue));
         
@@ -631,60 +700,43 @@ hints_setup(RADIUS_REQ *req)
                         avl_merge(&request_pairs, &tmp);
         }
 
-        for (i = hints; i; i = i->next) {
+	itr = iterator_create(hints);
+	for (rule = iterator_first(itr); rule; rule = iterator_next(itr)) {
                 int do_strip;
                 VALUE_PAIR *add;
                 
-                if (matches(req, name_pair->avp_strvalue, i, newname))
+                if (matches(req, name_pair->avp_strvalue, rule, newname))
                         continue;
 
                 matched++;
-                
-                debug(1, ("matched %s at hints:%d", i->name, i->lineno));
+                radius_req_register_locus(req, &rule->loc);
+		
+                debug(1, ("matched %s at %s:%lu", rule->name, rule->loc.file,
+			  (unsigned long) rule->loc.line));
         
-                add = avl_dup(i->rhs);
+                add = avl_dup(rule->rhs);
         
                 /* See if we need to adjust the name. */
                 do_strip = 1;
-                if ((tmp = avl_find(i->rhs, DA_STRIP_USER_NAME)) != NULL
-                    || (tmp = avl_find(i->lhs, DA_STRIP_USER_NAME)) != NULL)
+                if ((tmp = avl_find(rule->rhs, DA_STRIP_USER_NAME)) != NULL
+                    || (tmp = avl_find(rule->lhs, DA_STRIP_USER_NAME)) != NULL)
                         do_strip = tmp->avp_lvalue;
                 
                 if (do_strip) 
                         string_replace(&name_pair->avp_strvalue, newname);
 
-                /* Ok, let's see if we need to further modify the username */
-                if ((tmp = avl_find(i->rhs, DA_REPLACE_USER_NAME))
-                    || (tmp = avl_find(i->lhs, DA_REPLACE_USER_NAME))) {
-                        char *ptr;
-                        struct obstack hints_stk;
- 
-                        obstack_init(&hints_stk);
-                        ptr = radius_xlate(&hints_stk, tmp->avp_strvalue,
-                                           req, NULL);
-                        if (ptr) 
-                                string_replace(&name_pair->avp_strvalue, ptr);
-                        obstack_free(&hints_stk, NULL);
-                }
-                
-                /* Is the rewrite function specified? */
-                if ((tmp = avl_find(i->rhs, DA_REWRITE_FUNCTION))
-                    || (tmp = avl_find(i->lhs, DA_REWRITE_FUNCTION))) {
-                        if (run_rewrite(tmp->avp_strvalue, req)) {
-                                radlog(L_ERR, "hints:%d: %s(): %s",
-                                       i->lineno,
-                                       tmp->avp_strvalue,
-				       _("not defined"));
-                        }
-                }
-
+		/* Handle Rewrite-Function and Replace-User-Name */
+		hints_eval_compat(req, name_pair, rule);
+		/* Evaluate hints */
+		radius_eval_avl(req, add); /*FIXME: return value?*/
+		
                 debug(1, ("new name is `%s'", name_pair->avp_strvalue));
 
                 /* fix-up the string length */
                 name_pair->avp_strlength = strlen(name_pair->avp_strvalue);
 
                 /* Add all attributes to the request list, except
-                 * DA_STRIP_USER_NAME and DA_REPLACE_USER_NAME */
+		   DA_STRIP_USER_NAME and DA_REPLACE_USER_NAME */
                 avl_delete(&add, DA_STRIP_USER_NAME);
                 avl_delete(&add, DA_REPLACE_USER_NAME);
                 avl_delete(&add, DA_REWRITE_FUNCTION);
@@ -693,27 +745,29 @@ hints_setup(RADIUS_REQ *req)
                 
                 /* Ok, let's see if we need to further check the
                    hint's rules */
-                if (((tmp = avl_find(i->rhs, DA_FALL_THROUGH)) != NULL
-                     || (tmp = avl_find(i->lhs, DA_FALL_THROUGH)) != NULL)
+                if (((tmp = avl_find(rule->rhs, DA_FALL_THROUGH)) != NULL
+                     || (tmp = avl_find(rule->lhs, DA_FALL_THROUGH)) != NULL)
                     && tmp->avp_lvalue)
                         continue;
                 break;
         }
 
+	iterator_destroy(&itr);
+	
         if (matched) {
-		if (orig_name_pair)
-			avl_add_pair(&request_pairs, orig_name_pair);
-		else
-			avl_add_pair(&request_pairs, name_pair);
-	} else {
-		if (orig_name_pair)
-			avp_free(orig_name_pair);
-		else
-			avp_free(name_pair);
-	}
+                if (orig_name_pair)
+                        avl_add_pair(&request_pairs, orig_name_pair);
+                else
+                        avl_add_pair(&request_pairs, name_pair);
+        } else {
+                if (orig_name_pair)
+                        avp_free(orig_name_pair);
+                else
+                        avp_free(name_pair);
+        }
 
         req->request = request_pairs;
-	
+        
         return 0;
 }
 
@@ -727,19 +781,24 @@ hints_setup(RADIUS_REQ *req)
 int
 huntgroup_match(RADIUS_REQ *req, char *huntgroup)
 {
-        MATCHING_RULE *pl;
-        
-        for (pl = huntgroups; pl; pl = pl->next) {
-                if (strcmp(pl->name, huntgroup) != 0)
+        MATCHING_RULE *rule;
+	ITERATOR *itr;
+
+        if (!huntgroups)
+		return 0;
+	itr = iterator_create(huntgroups);
+	for (rule = iterator_first(itr); rule; rule = iterator_next(itr)) {
+                if (strcmp(rule->name, huntgroup) != 0)
                         continue;
-                if (paircmp(req, pl->lhs, NULL) == 0) {
-                        debug(1, ("matched %s at huntgroups:%d",
-                                 pl->name, pl->lineno));
+                if (paircmp(req, rule->lhs, NULL) == 0) {
+                        debug(1, ("matched %s at %s:%lu",
+				  rule->name, rule->loc.file,
+				  (unsigned long) rule->loc.line));
                         break;
                 }
         }
-
-        return (pl != NULL);
+	iterator_destroy(&itr);
+        return (rule != NULL);
 }
 
 
@@ -750,38 +809,43 @@ huntgroup_match(RADIUS_REQ *req, char *huntgroup)
  *         -1 on error.
  */
 int
-huntgroup_access(RADIUS_REQ *radreq)
+huntgroup_access(RADIUS_REQ *radreq, LOCUS *loc)
 {
         VALUE_PAIR      *pair;
-        MATCHING_RULE   *pl;
+        MATCHING_RULE   *rule;
+	ITERATOR *itr;
         int             r = 1;
 
         if (huntgroups == NULL)
                 return 1;
 
-        for (pl = huntgroups; pl; pl = pl->next) {
+	itr = iterator_create(huntgroups);
+	for (rule = iterator_first(itr); rule; rule = iterator_next(itr)) {
                 /*
                  *      See if this entry matches.
                  */
-                if (paircmp(radreq, pl->lhs, NULL) != 0)
+                if (paircmp(radreq, rule->lhs, NULL) != 0)
                         continue;
-                debug(1, ("matched huntgroup at huntgroups:%d", pl->lineno));
-                r = paircmp(radreq, pl->rhs, NULL) == 0;
+                debug(1, ("matched huntgroup at %s:%lu", rule->loc.file,
+			  (unsigned long) rule->loc.line));
+                r = paircmp(radreq, rule->rhs, NULL) == 0;
                 break;
         }
-
+	iterator_destroy(&itr);
+	
+        if (rule) {
+		if (loc)
+			*loc = rule->loc;
+		radius_req_register_locus(radreq, &rule->loc);
 #ifdef DA_REWRITE_FUNCTION
-        if (pl &&
-            (pair = avl_find(pl->lhs, DA_REWRITE_FUNCTION)) != NULL) {
-                if (run_rewrite(pair->avp_strvalue, radreq)) {
-                        radlog(L_ERR, "huntgroups:%d: %s(): %s",
-                               pl->lineno,
-                               pair->avp_strvalue,
-			       _("not defined"));
+		if ((pair = avl_find(rule->lhs, DA_REWRITE_FUNCTION)) != NULL
+		    && rewrite_eval(pair->avp_strvalue, radreq, NULL, NULL)) {
+                        radlog_loc(L_ERR, &rule->loc, "%s(): %s",
+				   pair->avp_strvalue,
+				   _("not defined"));
                 }
-        }
 #endif  
-
+	}
         debug(1, ("returning %d", r));
         return r;
 }
@@ -804,26 +868,26 @@ read_naslist_file(char *file)
  */
 /*ARGSUSED*/
 int
-read_clients_entry(void *u ARG_UNUSED, int fc, char **fv,
-		   char *file, int lineno)
+read_clients_entry(void *u ARG_UNUSED, int fc, char **fv, LOCUS *loc)
 {
         CLIENT *cp;
         
         if (fc != 2) {
-                radlog(L_ERR, "%s:%d: %s",
-                       file, lineno,
-		       _("wrong number of fields"));
+                radlog_loc(L_ERR, loc, "%s", _("wrong number of fields"));
                 return -1;
         }
 
         cp = emalloc(sizeof(CLIENT));
 
-        cp->ipaddr = ip_gethostaddr(fv[0]);
+	if (strcmp(fv[0], "DEFAULT") == 0) 
+		cp->netdef.ipaddr = cp->netdef.netmask = 0;
+	else
+		ip_getnetaddr(fv[0], &cp->netdef);
         cp->secret = estrdup(fv[1]);
         if (fc == 3)
                 STRING_COPY(cp->shortname, fv[2]);
-        ip_gethostname(cp->ipaddr, cp->longname, sizeof(cp->longname));
-	list_append(clients, cp);
+        ip_gethostname(cp->netdef.ipaddr, cp->longname, sizeof(cp->longname));
+        list_append(clients, cp);
         return 0;
 }
 
@@ -831,9 +895,9 @@ static int
 client_free(void *item, void *data ARG_UNUSED)
 {
         CLIENT *cl = item;
-	efree(cl->secret);
-	efree(cl);
-	return 0;
+        efree(cl->secret);
+        efree(cl);
+        return 0;
 }
 
 /*
@@ -842,8 +906,8 @@ client_free(void *item, void *data ARG_UNUSED)
 int
 read_clients_file(char *file)
 {
-	list_destroy(&clients, client_free, NULL);
-	clients = list_create();
+        list_destroy(&clients, client_free, NULL);
+        clients = list_create();
         return read_raddb_file(file, 1, read_clients_entry, NULL);
 }
 
@@ -855,14 +919,14 @@ CLIENT *
 client_lookup_ip(UINT4 ipaddr)
 {
         CLIENT *cl;
-	ITERATOR *itr = iterator_create(clients);
+        ITERATOR *itr = iterator_create(clients);
 
-	if (!itr)
-		return NULL;
+        if (!itr)
+                return NULL;
         for (cl = iterator_first(itr); cl; cl = iterator_next(itr))
-                if (ipaddr == cl->ipaddr)
+		if (ip_addr_in_net_p(&cl->netdef, ipaddr))
                         break;
-	iterator_destroy(&itr);
+        iterator_destroy(&itr);
         return cl;
 }
 
@@ -894,15 +958,13 @@ client_lookup_name(UINT4 ipaddr, char *buf, size_t bufsize)
  */
 /*ARGSUSED*/
 int
-read_nastypes_entry(void *u ARG_UNUSED, int fc, char **fv,
-		    char *file, int lineno)
+read_nastypes_entry(void *u ARG_UNUSED, int fc, char **fv, LOCUS *loc)
 {
         RADCK_TYPE *mp;
         int method;
 
         if (fc < 2) {
-                radlog(L_ERR, "%s:%d: %s", file, lineno,
-		       _("too few fields"));
+                radlog_loc(L_ERR, loc, "%s", _("too few fields"));
                 return -1;
         }
 
@@ -913,7 +975,7 @@ read_nastypes_entry(void *u ARG_UNUSED, int fc, char **fv,
         else if (strcmp(fv[1], "ext") == 0)
                 method = METHOD_EXT;
         else {
-                radlog(L_ERR, "%s:%d: %s", file, lineno, _("unknown method"));
+                radlog_loc(L_ERR, loc, "%s", _("unknown method"));
                 return -1;
         }
                         
@@ -922,7 +984,7 @@ read_nastypes_entry(void *u ARG_UNUSED, int fc, char **fv,
         mp->method = method;
         if (fc > 2)
                 mp->args = envar_parse_argcv(fc-2, &fv[2]);
-        else
+	else
                 mp->args = NULL;
 	list_append(radck_type, mp);
         return 0;
@@ -979,19 +1041,17 @@ add_deny(char *user)
 
 /*ARGSUSED*/
 int
-read_denylist_entry(int *denycnt, int fc, char **fv, char *file, int lineno)
+read_denylist_entry(void *closure, int fc, char **fv, LOCUS *loc)
 {
-        if (fc != 1) {
-                radlog(L_ERR,
-		       "%s:%d: %s",
-                       file, lineno,
-		       _("wrong number of fields"));
+	int *denycnt = closure;
+	if (fc != 1) {
+                radlog_loc(L_ERR, loc, "%s", _("wrong number of fields"));
                 return -1;
         }
 
         if (get_deny(fv[0]))
-                radlog(L_ERR, _("user `%s' already found in %s"),
-                    fv[0], RADIUS_DENY);
+                radlog_loc(L_ERR, loc, _("user `%s' already found in %s"),
+			   fv[0], RADIUS_DENY);
         else {
                 add_deny(fv[0]);
                 (*denycnt)++;
@@ -1224,7 +1284,7 @@ paircmp(RADIUS_REQ *request, VALUE_PAIR *check, char *pusername)
                 if (debug_on(20)) {
                         radlog(L_DEBUG, 
                                "check_item: %s", 
-                               format_pair(check_item, &save));
+                               format_pair(check_item, 1, &save));
                         free(save);
                 }
 
@@ -1271,7 +1331,7 @@ paircmp(RADIUS_REQ *request, VALUE_PAIR *check, char *pusername)
                 if (debug_on(20)) {
                         radlog(L_DEBUG,
                                "auth_item: %s",
-                               format_pair(auth_item, &save));
+                               format_pair(auth_item, 1, &save));
                         free(save);
                 }
 
@@ -1346,22 +1406,18 @@ paircmp(RADIUS_REQ *request, VALUE_PAIR *check, char *pusername)
 /*
  * Free a MATCHING_RULE
  */
-void
-matchrule_free(MATCHING_RULE **pl)
+static int
+matching_rule_free(void *item, void *data ARG_UNUSED)
 {
-        MATCHING_RULE *p, *next;
-
-        for (p = *pl; p; p = next) {
-                if (p->name)
-                        efree(p->name);
-                if (p->lhs)
-                        avl_free(p->lhs);
-                if (p->rhs)
-                        avl_free(p->rhs);
-                next = p->next;
-                efree(p);
-        }
-        *pl = NULL;
+	MATCHING_RULE *p = item;
+	
+	if (p->name)
+		efree(p->name);
+	if (p->lhs)
+		avl_free(p->lhs);
+	if (p->rhs)
+		avl_free(p->rhs);
+	efree(p);
 }
 
 /* ***************************************************************************
@@ -1543,7 +1599,8 @@ reload_data(enum reload_what what, int *do_radck)
         
         switch (what) {
         case reload_all:
-                /* This implies reloading users, huntgroups and hints */
+                /* This implies reloading rewrite, users, huntgroups 
+		   and hints */
                 rc += reload_data(reload_dict, do_radck);
                 rc += reload_data(reload_clients, do_radck);
                 rc += reload_data(reload_naslist, do_radck);
@@ -1552,7 +1609,6 @@ reload_data(enum reload_what what, int *do_radck)
 #ifdef USE_SQL
                 rc += reload_data(reload_sql, do_radck);
 #endif
-                reload_data(reload_rewrite, do_radck);
                 break;
                 
         case reload_users:
@@ -1593,23 +1649,20 @@ reload_data(enum reload_what what, int *do_radck)
                 if (dict_init())
                         rc = 1;
 
-                /* Users, huntgroups and hints should be reloaded after
-                 * changing dictionary.
-                 */
-                rc += reload_data(reload_users, do_radck);
-                rc += reload_data(reload_huntgroups, do_radck);
-                rc += reload_data(reload_hints, do_radck);
+                /* reload_rewrite implies reloading users, huntgroups
+		   and hints */
+                rc += reload_data(reload_rewrite, do_radck);
                 break;
                 
         case reload_huntgroups:
-                matchrule_free(&huntgroups);
+                list_destroy(&huntgroups, matching_rule_free, NULL);
                 path = mkfilename(radius_dir, RADIUS_HUNTGROUPS);
                 huntgroups = file_read(CF_HUNTGROUPS, path);
                 efree(path);
                 break;
                 
         case reload_hints:
-                matchrule_free(&hints);
+                list_destroy(&hints, matching_rule_free, NULL);
                 path = mkfilename(radius_dir, RADIUS_HINTS);
                 hints = file_read(CF_HINTS, path);
                 efree(path);
@@ -1626,11 +1679,11 @@ reload_data(enum reload_what what, int *do_radck)
 
         case reload_naslist:
                 /*FIXME*/
-                path = mkfilename(radius_dir, RADIUS_NASTYPES);
+		path = mkfilename(radius_dir, RADIUS_NASTYPES);
                 read_nastypes_file(path);
                 efree(path);
                 /*EMXIF*/
-                
+
                 path = mkfilename(radius_dir, RADIUS_NASLIST);
                 if (read_naslist_file(path) < 0)
                         rc = 1;
@@ -1660,9 +1713,10 @@ reload_data(enum reload_what what, int *do_radck)
 #endif
 
         case reload_rewrite:
-                path = mkfilename(radius_dir, "rewrite");
-                rc = parse_rewrite(path);
-                efree(path);
+		rewrite_load_all();
+                rc += reload_data(reload_users, do_radck);
+                rc += reload_data(reload_huntgroups, do_radck);
+                rc += reload_data(reload_hints, do_radck);
                 break;
                 
         default:
@@ -1688,78 +1742,25 @@ reload_config_file(enum reload_what what)
 /* ****************************************************************************
  * Debugging functions
  */
-static void
-dump_pairs(FILE *fp, VALUE_PAIR *pair)
-{
-        int etype;
-        
-        for (; pair; pair = pair->next) {
-                fprintf(fp, "\t\t%s %s ", pair->name, 
-                        op_tab[pair->operator].name);
-
-                switch (pair->type) {
-                case TYPE_STRING:
-                        fprintf(fp, "(STRING) ");
-                        break;
-
-                case TYPE_INTEGER:
-                        fprintf(fp, "(INTEGER) ");
-                        break;
-
-                case TYPE_IPADDR:
-                        fprintf(fp, "(IP) ");
-                        break;
-                
-                case TYPE_DATE:
-                        fprintf(fp, "(DATE) ");
-                        break;
-                        
-                default:
-                        fprintf(fp, "(%d) ", pair->type);
-                }
-
-                if (pair->eval) {
-                        etype = TYPE_STRING;
-                        fprintf(fp, "=");
-                } else
-                        etype = pair->type;
-                
-                switch (etype) {
-                case TYPE_STRING:
-                        fprintf(fp, "%s", pair->avp_strvalue);
-                        break;
-
-                case TYPE_INTEGER:
-                        fprintf(fp, "%ld", pair->avp_lvalue);
-                        break;
-
-                case TYPE_IPADDR:
-                        fprintf(fp, "%lx", pair->avp_lvalue);
-                        break;
-                
-                case TYPE_DATE:
-                        fprintf(fp, "%ld", pair->avp_lvalue);
-                        break;
-                        
-                }
-                fprintf(fp, "\n");
-        }
-}
-
 void
-dump_pair_list(FILE *fp, char *header, MATCHING_RULE *pl)
+dump_matching_rules(FILE *fp, char *header, RAD_LIST *list)
 {
+	MATCHING_RULE *rule;
+	ITERATOR *itr = iterator_create(list);
+
         fprintf(fp, "%s {\n", header);
-        for ( ; pl; pl = pl->next) {
-                fprintf(fp, "\t%s:\n", pl->name);
+ 	for (rule = iterator_first(itr); rule; rule = iterator_next(itr)) {
+                fprintf(fp, "\t%s:\n", rule->name);
                 fprintf(fp, "\tlhs {\n");
-                dump_pairs(fp, pl->lhs);
+		avl_fprint(fp, "\t\t", 1, rule->lhs);
                 fprintf(fp, "\t}\n");
 
-                fprintf(fp, "\trhs {\n");
-                dump_pairs(fp, pl->rhs);
+                fprintf(fp, "\trhs {\n"); 
+		avl_fprint(fp, "\t\t", 1, rule->rhs);
                 fprintf(fp, "\t}\n");
         }
+	iterator_destroy(&itr);
+	
         fprintf(fp, "}\n");
 }
 
@@ -1768,11 +1769,11 @@ dump_user(FILE *fp, User_symbol *sym)
 {
         fprintf(fp, "\t%s:\n", sym->name);
         fprintf(fp, "\tlhs {\n");
-        dump_pairs(fp, sym->check);
+	avl_fprint(fp, "\t\t", 1, sym->check);
         fprintf(fp, "\t}\n");
 
         fprintf(fp, "\trhs {\n");
-        dump_pairs(fp, sym->reply);
+	avl_fprint(fp, "\t\t", 1, sym->reply);
         fprintf(fp, "\t}\n");
         
         return 0;
@@ -1799,8 +1800,8 @@ dump_users_db()
         symtab_iterate(user_tab, dump_user, fp);
         fprintf(fp, "}\n");
 
-        dump_pair_list(fp, "huntgroups", huntgroups);
-        dump_pair_list(fp, "hints", hints);
+        dump_matching_rules(fp, "huntgroups", huntgroups);
+        dump_matching_rules(fp, "hints", hints);
         radlog(L_INFO, _("dumped users database to %s"), RADIUS_DUMPDB_NAME);
         fclose(fp);
         efree(name);
@@ -1839,16 +1840,3 @@ strip_username(int do_strip, char *name, VALUE_PAIR *check_item,
 
 
 
-
-
-
-
-
-
-
-
-
-
-
-
-
