@@ -1,5 +1,7 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000, 2002, 2003, Sergey Poznyakoff
+   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
   
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,6 +24,7 @@
 #endif
 
 #include <radiusd.h>
+#include <obstack1.h>
 
 /* Build and send a reply to the incoming request.
    Input: code        -- Reply code.
@@ -48,17 +51,25 @@ radius_send_reply(int code, RADIUS_REQ *radreq,
                 
                 switch (code) {
                 case RT_PASSWORD_REJECT:
-                case RT_AUTHENTICATION_REJECT:
+                case RT_ACCESS_REJECT:
                         radreq->reply_pairs = NULL;
                         avl_move_attr(&radreq->reply_pairs, &reply, 
                                       DA_REPLY_MESSAGE);
                         avl_move_attr(&radreq->reply_pairs, &reply, 
                                       DA_PROXY_STATE);
                         avl_free(reply);
+			stat_inc(auth, radreq->ipaddr, num_rejects);
                         break;
 
+		case RT_ACCESS_ACCEPT:
+			stat_inc(auth, radreq->ipaddr, num_accepts);
+			/*FALLTHROUGH*/
+			
                 default:
-                        radreq->reply_pairs = reply;
+                        radreq->reply_pairs =
+				rad_clt_encrypt_pairlist(reply,
+							 radreq->vector,
+							 radreq->secret);
                 }
         } 
 
@@ -162,7 +173,7 @@ radius_auth_req_decode(struct sockaddr_in *sa,
 		radreq_free(radreq);
  		return 1;
         }
-
+	
 	/* RFC 2865 p. 2.2:
 	   The random challenge can either be included in the
 	   CHAP-Challenge attribute or, if it is 16 octets long,
@@ -211,39 +222,70 @@ radius_acct_req_decode(struct sockaddr_in *sa,
 	return 0;
 }
 
+static void
+decrypt_pair(RADIUS_REQ *req, VALUE_PAIR *pair)
+{
+	if (pair->prop & AP_ENCRYPT) {
+		char password[AUTH_STRING_LEN+1];
+		req_decrypt_password(password, req, pair);
+		efree(pair->avp_strvalue);
+		pair->avp_strvalue = estrdup(password);
+		pair->avp_strlength = strlen(pair->avp_strvalue);
+	}
+}
+
+VALUE_PAIR *
+radius_decrypt_request_pairs(RADIUS_REQ *req, VALUE_PAIR *plist)
+{
+	VALUE_PAIR *pair;
+
+	for (pair = plist; pair; pair = pair->next) 
+		decrypt_pair(req, pair);
+	
+	return plist;
+}
+
+void
+radius_destroy_pairs(VALUE_PAIR **p)
+{
+	VALUE_PAIR *pair;
+	
+	if (!p || !*p)
+		return;
+	for (pair = *p; pair; pair = pair->next) {
+		if (pair->prop & (AP_ENCRYPT_RFC2138|AP_ENCRYPT_RFC2868))
+			memset(pair->avp_strvalue, 0,
+			       pair->avp_strlength);
+	}
+
+	avl_free(*p);
+	*p = NULL;
+}
+
 static VALUE_PAIR *
 _extract_pairs(RADIUS_REQ *req, int prop)
 {
-	static int attrlist[] = { DA_USER_PASSWORD, DA_CHAP_PASSWORD };
 	int i;
 	VALUE_PAIR *newlist = NULL;
 	VALUE_PAIR *pair;
 	char password[AUTH_STRING_LEN+1];
 	int found = 0;
 	
-	for (pair = req->request; !found && pair; pair = pair->next) 
-		for (i = 0; i < NITEMS(attrlist); i++) {
-			if (pair->attribute == attrlist[i]
-			    && (pair->prop & prop)) {
-				found = 1;
-				break;
-			}
+	for (pair = req->request; !found && pair; pair = pair->next)
+		if (pair->prop &
+		    (prop|AP_ENCRYPT_RFC2138|AP_ENCRYPT_RFC2868)) {
+			found = 1;
+			break;
 		}
 
 	if (!found)
 		return NULL;
 
 	newlist = avl_dup(req->request);
-	for (pair = newlist; pair; pair = pair->next) 
-		for (i = 0; i < NITEMS(attrlist); i++) {
-			if (pair->attribute == attrlist[i]
-			    && (pair->prop & prop)) {
-				req_decrypt_password(password, req, pair);
-				efree(pair->avp_strvalue);
-				pair->avp_strvalue = estrdup(password);
-				pair->avp_strlength = strlen(pair->avp_strvalue);
-			}
-		}
+	for (pair = newlist; pair; pair = pair->next) { 
+		if (pair->prop & prop) 
+			decrypt_pair(req, pair);
+	}
 	return newlist;
 }
 
@@ -278,9 +320,9 @@ radius_req_cmp(void *adata, void *bdata)
 	nas = nas_request_to_nas(a);
 
 	switch (a->code) {
-	case RT_AUTHENTICATION_REQUEST:
-	case RT_AUTHENTICATION_ACK:
-	case RT_AUTHENTICATION_REJECT:
+	case RT_ACCESS_REQUEST:
+	case RT_ACCESS_ACCEPT:
+	case RT_ACCESS_REJECT:
 	case RT_ACCESS_CHALLENGE:
 		prop = find_prop(nas, "compare-auth-flag", 0);
 		if (!prop)
@@ -312,8 +354,8 @@ radius_req_cmp(void *adata, void *bdata)
 	
 	rc = avl_cmp(ap, bp, prop) || avl_cmp(bp, ap, prop);
 
-	avl_free(alist);
-	avl_free(blist);
+	radius_destroy_pairs(&alist);
+	radius_destroy_pairs(&blist);
 	
 	if (rc == 0) {
 		/* We need to replace A/V pairs, authenticator and ID
@@ -347,9 +389,7 @@ radius_req_update(void *req_ptr, void *data_ptr)
 void
 radius_req_free(void *req)
 {
-        debug(1,("enter: %p",req));
         radreq_free((RADIUS_REQ *)req);
-        debug(1,("exit"));
 }
 
 /*ARGSUSED*/
@@ -406,34 +446,52 @@ radius_req_failure(int type, struct sockaddr_in *addr)
 	return 0;
 }
 
+int
+radius_status_server(RADIUS_REQ *radreq, int fd)
+{
+	radius_send_reply(RT_ACCESS_ACCEPT, radreq, NULL,
+			  "GNU Radius server fully operational",
+			  fd);
+	return 0;
+}
+
 int
 radius_respond(REQUEST *req)
 {
 	int rc;
 	RADIUS_REQ *radreq = req->data;
-	
+
+	forward_request(req->type, radreq);
+
 #ifdef USE_SQL
 	radiusd_sql_clear_cache();
 #endif
+
+	if (radreq->code == RT_ACCESS_REQUEST
+	    && rad_auth_check_username(radreq, req->fd))
+	    return 1;
+	
         /* Add any specific attributes for this username. */
         hints_setup(radreq);
 
         /* Check if we support this request */
         switch (radreq->code) {
-        case RT_AUTHENTICATION_REQUEST:
+        case RT_ACCESS_REQUEST:
                 stat_inc(auth, radreq->ipaddr, num_access_req);
                 if (rad_auth_init(radreq, req->fd) < 0) 
                         return 1;
-		/*FALLTHRU*/
-		
-        case RT_ACCOUNTING_REQUEST:
-                if (proxy_send(req) != 0) {
+                if (proxy_send(req) != 0) 
                         return 0;
-                }
 		break;
 		
-	case RT_AUTHENTICATION_ACK:
-	case RT_AUTHENTICATION_REJECT:
+        case RT_ACCOUNTING_REQUEST:
+		stat_inc(acct, radreq->ipaddr, num_req);
+                if (proxy_send(req) != 0) 
+                        return 0;
+		break;
+		
+	case RT_ACCESS_ACCEPT:
+	case RT_ACCESS_REJECT:
 	case RT_ACCOUNTING_RESPONSE:
 	case RT_ACCESS_CHALLENGE:
 		if (!req->orig) {
@@ -453,7 +511,7 @@ radius_respond(REQUEST *req)
 	}
 
         switch (radreq->code) {
-        case RT_AUTHENTICATION_REQUEST:
+        case RT_ACCESS_REQUEST:
 		rad_authenticate(radreq, req->fd);
                 break;
 		
@@ -464,12 +522,113 @@ radius_respond(REQUEST *req)
 			stat_inc(acct, radreq->ipaddr, num_bad_sign);
 		rad_accounting(radreq, req->fd, rc);
                 break;
-		
+
+	case RT_STATUS_SERVER:
+		radius_status_server(radreq, req->fd);
+		break;
+					
         default:
                 stat_inc(acct, radreq->ipaddr, num_unknowntypes);
                 radlog_req(L_NOTICE, radreq, _("unknown request code %d"), 
                            radreq->code); 
                 return -1;
         }       
+
+	radius_trace_path(radreq);
+	
         return 0;
 }
+
+void
+radius_req_register_locus(RADIUS_REQ *req, LOCUS *loc)
+{
+	switch (req->code) {
+	case RT_ACCESS_REQUEST:
+	case RT_ACCESS_ACCEPT:
+	case RT_ACCESS_REJECT:
+	case RT_ACCESS_CHALLENGE:
+		if (!auth_trace_rules)
+			return;
+		break;
+		
+	case RT_ACCOUNTING_REQUEST:
+	case RT_ACCOUNTING_RESPONSE:
+	case RT_ACCOUNTING_STATUS:
+	case RT_ACCOUNTING_MESSAGE:
+		if (!acct_trace_rules)
+			return;
+		break;
+
+	default:
+		return;
+	}
+
+	if (!req->locus_list)
+		req->locus_list = list_create();
+	
+	list_prepend(req->locus_list, loc);
+}
+
+struct trace_data {
+	struct obstack stk;
+	char *file;
+};
+
+static char *
+skip_common_substring(char *str, char *pat)
+{
+	char *start = str;
+
+	while (*str == *pat++) 
+		if (*str++ == '/')
+			start = str;
+	return start; 
+}
+
+static int
+_trace_path_compose(void *item, void *data)
+{
+	LOCUS *loc = item;
+	struct trace_data *td = data;
+	char buf[64];
+	
+	if (!td->file) {
+		td->file = loc->file;
+		obstack_grow(&td->stk, loc->file, strlen(loc->file));
+		obstack_1grow(&td->stk, ':');
+	} else if (strcmp(td->file, loc->file) == 0) {
+		obstack_1grow(&td->stk, ',');
+	} else {
+		char *p;
+		
+		obstack_1grow(&td->stk, ';');
+		obstack_1grow(&td->stk, ' ');
+		
+		p = skip_common_substring(loc->file, td->file);
+		obstack_grow(&td->stk, p, strlen(p));
+		obstack_1grow(&td->stk, ':');
+		td->file = loc->file;
+	}
+
+	snprintf(buf, sizeof buf, "%lu", (unsigned long) loc->line);
+	obstack_grow(&td->stk, buf, strlen(buf));
+	
+	return 0;
+}
+
+void
+radius_trace_path(RADIUS_REQ *req)
+{
+	struct trace_data td;
+	char *p;
+	
+	if (!req->locus_list)
+		return; 
+	
+	obstack_init(&td.stk);
+	td.file = NULL;
+	list_iterate(req->locus_list, _trace_path_compose, &td);
+	p = obstack_finish(&td.stk);
+	radlog_req(L_INFO, req, _("rule trace: %s"), p);
+	obstack_free(&td.stk, NULL);
+}
