@@ -87,17 +87,27 @@ static struct argp_option options[] = {
 
 /* *************************** Global Variables **************************** */
 
-int        debug_flag; /* can be raised from debugger only */
-int        log_mode;
+int        debug_flag;     /* can be raised from debugger only */
+int        log_mode;       /* logging mode */
+char       *auth_log_hook; /* Authentication logging hook function */
 
 static int foreground; /* Stay in the foreground */
-int spawn_flag; 
+int spawn_flag;        /* Whether to spawn new children for handling
+			  the requests */
 
-int use_dbm = 0;
-int auth_detail = 0;
-int acct_detail = 1;      
-int strip_names;
-int suspend_flag;
+int use_dbm = 0;       /* Use DBM storage */
+int auth_detail = 0;   /* Produce detailed logs of authentication packets */
+int acct_detail = 1;   /* Produce detailed logs of accounting packets */
+int auth_trace_rules = 0; /* Produce trace logs for each auth request */
+int acct_trace_rules = 0; /* Produce trace logs for each acct request */
+int strip_names;          /* Strip preffixes/suffixes off the usernames */
+int suspend_flag;         /* Suspend processing of RADIUS requests */
+int auth_reject_malformed_names = 0; /* Respond with Access-Reject packets
+					for requests with malformed user
+					names */
+
+RADIUS_USER radiusd_user; /* Run the daemon with this user privileges */
+RADIUS_USER exec_user;    /* Run the user programs with this user privileges */
 
 #define CMD_NONE     0 /* No command */
 #define CMD_CLEANUP  1 /* Cleanup finished children */
@@ -129,7 +139,6 @@ size_t max_children = MAX_CHILDREN;
 unsigned process_timeout = PROCESS_TIMEOUT;
 unsigned radiusd_write_timeout = RADIUSD_WRITE_TIMEOUT;
 unsigned radiusd_read_timeout = RADIUSD_READ_TIMEOUT;
-char *exec_user = NULL;
 
 UINT4 warning_seconds;
 int use_guile;
@@ -270,7 +279,6 @@ static struct argp argp = {
 void
 set_config_defaults()
 {
-        exec_user  = estrdup("daemon");
         username_valid_chars = estrdup(".-_!@#$%^&\\/");
         message_text[MSG_ACCOUNT_CLOSED] =
                 estrdup(_("Sorry, your account is currently closed\n"));
@@ -360,9 +368,14 @@ radiusd_preconfig_hook(void *a ARG_UNUSED, void *b ARG_UNUSED)
 static void
 radiusd_postconfig_hook(void *a ARG_UNUSED, void *b ARG_UNUSED)
 {
-	if (radius_mode = MODE_DAEMON && radius_count_channels() == 0) {
-		radlog(L_ALERT,
-		       _("Radiusd is not listening on any port. Trying to continue anyway..."));
+	if (radius_mode == MODE_DAEMON && radius_count_channels() == 0) {
+		if (foreground) {
+			radlog(L_ALERT,
+			       _("Radiusd is not listening on any port."));
+			exit(1);
+		} else
+			radlog(L_ALERT,
+			       _("Radiusd is not listening on any port. Trying to continue anyway..."));
 	}
 }
 
@@ -388,7 +401,8 @@ radiusd_setup()
 	   executed after all other hooks */
 	radiusd_set_preconfig_hook(radiusd_preconfig_hook, NULL, 0);
 	radiusd_set_postconfig_hook(radiusd_postconfig_hook, NULL, 0);
-	
+
+	rewrite_init();
         snmp_init(0, 0, (snmp_alloc_t)emalloc, (snmp_free_t)efree);
 }
 
@@ -411,11 +425,13 @@ common_init()
         setvbuf(stdout, NULL, _IOLBF, 0);
 #endif
 	radiusd_signal_init(sig_handler);
-	radiusd_reconfigure();
-	radpath_init();
+	forward_init();
 #ifdef USE_SNMP
         snmpserv_init(&saved_status);
-#endif		
+#endif
+	acct_init();
+	radiusd_reconfigure();
+	radpath_init();
 	radlog(L_INFO, _("Ready"));
 }
 
@@ -443,7 +459,7 @@ radiusd_daemon()
 #endif
         /* SIGHUP is ignored because when the session leader terminates
            all process in the session are sent the SIGHUP.  */
-        signal(SIGHUP, SIG_IGN);
+        rad_set_signal(SIGHUP, SIG_IGN);
 
         /* fork() again so the parent, can exit. This means that we, as a
            non-session group leader, can never regain a controlling
@@ -472,6 +488,7 @@ radiusd_daemon()
                 fflush(stdout);
                 fflush(stderr);
         }
+	
         efree(p);
 }
 
@@ -521,6 +538,15 @@ radiusd_main()
 
 	radiusd_pidfile_write(RADIUSD_PID_FILE);
 
+	if (radiusd_user.username) {
+		char *p;
+		log_change_owner(&radiusd_user);
+		p = mkfilename(radlog_dir, "radius.stderr");
+		chown(p, radiusd_user.uid, radiusd_user.gid);
+		efree(p);
+		radius_switch_to_user(&radiusd_user);
+	}
+
         radiusd_main_loop();
 }
 
@@ -807,8 +833,8 @@ struct hook_rec {
 	int once; /* Run once and remove */
 };
 
-static LIST /* of struct hook_rec */ *preconfig;
-static LIST /* of struct hook_rec */ *postconfig;
+static RAD_LIST /* of struct hook_rec */ *preconfig;
+static RAD_LIST /* of struct hook_rec */ *postconfig;
 
 void
 radiusd_set_preconfig_hook(void (*f)(void *, void *), void *p, int once)
@@ -835,7 +861,7 @@ radiusd_set_postconfig_hook(void (*f)(void *, void *), void *p, int once)
 }
 
 struct hook_runtime_closure {
-	LIST *list;
+	RAD_LIST *list;
 	void *call_data;
 };
 
@@ -931,7 +957,7 @@ sig_handler(int sig)
 	default:
 		abort();
 	}
-	signal(sig, sig_handler);
+	rad_reset_signal(sig, sig_handler);
 }
 
 void
@@ -945,7 +971,7 @@ radiusd_signal_init(RETSIGTYPE (*hp)(int sig))
 	int i;
 
 	for (i = 0; i < sizeof(signum)/sizeof(signum[0]); i++)
-		signal(signum[i], hp);
+		rad_set_signal(signum[i], hp);
 }
 
 
@@ -1115,6 +1141,12 @@ rad_cfg_listen_auth(int argc, cfg_value_t *argv,
 		    void *block_data, void *handler_data)
 {
 	int i, errcnt = 0;
+
+	if (argc == 2 && argv[1].type == CFG_BOOLEAN) {
+		if (argv[1].v.bool == 0)
+			auth_port = 0;
+		return 0;
+	}
 	
 	for (i = 1; i < argc; i++)  
 		if (argv[i].type != CFG_HOST) {
@@ -1141,7 +1173,9 @@ auth_stmt_begin(int finish, void *block_data, void *handler_data)
 {
 	if (!finish) 
 		_opened_auth_sockets = 0;
-	else if (radius_mode == MODE_DAEMON && !_opened_auth_sockets) 
+	else if (radius_mode == MODE_DAEMON
+		 && !_opened_auth_sockets
+		 && auth_port)
 		udp_open(R_AUTH, INADDR_ANY, auth_port, 0);
 	return 0;
 }
@@ -1152,6 +1186,12 @@ rad_cfg_listen_acct(int argc, cfg_value_t *argv,
 {
 	int i, errcnt = 0;
 	
+	if (argc == 2 && argv[1].type == CFG_BOOLEAN) {
+		if (argv[1].v.bool == 0)
+			acct_port = 0;
+		return 0;
+	}
+	
 	for (i = 1; i < argc; i++)  
 		if (argv[i].type != CFG_HOST) {
 			cfg_type_error(CFG_HOST);
@@ -1175,11 +1215,37 @@ acct_stmt_begin(int finish, void *block_data, void *handler_data)
 {
 	if (!finish) 
 		_opened_acct_sockets = 0;
-	else if (radius_mode == MODE_DAEMON && !_opened_acct_sockets) 
+	else if (radius_mode == MODE_DAEMON
+		 && !_opened_acct_sockets
+		 && acct_port)
 		udp_open(R_ACCT, INADDR_ANY, acct_port, 0);
 	return 0;
 }
 
+static int
+rad_cfg_user(int argc, cfg_value_t *argv,
+	     void *block_data, void *handler_data)
+{
+	RADIUS_USER *usr = handler_data;
+
+	if (argc != 2 || argv[1].type != CFG_STRING) 
+		return 1;
+	return radius_get_user_ids((RADIUS_USER *) handler_data,
+				   argv[1].v.string);
+}
+
+int
+option_stmt_end(void *block_data, void *handler_data)
+{
+	if (radiusd_user.username && radiusd_user.uid != 0) {
+		radlog(L_WARN, _("Ignoring exec-program-user"));
+		efree(exec_user.username);
+		exec_user.username = NULL;
+	} else if (exec_user.username == NULL)
+		radius_get_user_ids(&exec_user, "daemon");
+	return 0;
+}
+
 struct cfg_stmt option_stmt[] = {
 	{ "source-ip", CS_STMT, NULL, cfg_get_ipaddr, &myip,
 	  NULL, NULL },
@@ -1195,8 +1261,10 @@ struct cfg_stmt option_stmt[] = {
 	  cfg_get_integer, &radiusd_read_timeout, NULL, NULL },
 	{ "master-write-timeout", CS_STMT, NULL,
 	  cfg_get_integer, &radiusd_write_timeout, NULL, NULL },
-	{ "exec-program-user", CS_STMT, NULL, cfg_get_string, &exec_user,
-	  NULL, NULL },
+	{ "exec-program-user", CS_STMT, NULL, rad_cfg_user,
+	  &exec_user, NULL, NULL },
+	{ "radiusd-user", CS_STMT, NULL, rad_cfg_user,
+	  &radiusd_user, NULL, NULL },
 	{ "log-dir", CS_STMT, NULL, cfg_get_string, &radlog_dir,
 	  NULL, NULL },
 	{ "acct-dir", CS_STMT, NULL, cfg_get_string, &radacct_dir,
@@ -1204,8 +1272,7 @@ struct cfg_stmt option_stmt[] = {
 	{ "resolve", CS_STMT, NULL, cfg_get_boolean, &resolve_hostnames,
 	  NULL, NULL },
 	{ "username-chars", CS_STMT, NULL, cfg_get_string,
-	  &username_valid_chars,
-	  NULL, NULL },
+	  &username_valid_chars, NULL, NULL },
 	/* Obsolete statements */
 	{ "usr2delay", CS_STMT, NULL, cfg_obsolete, NULL, NULL, NULL },
 	{ NULL, }
@@ -1242,6 +1309,7 @@ struct cfg_stmt message_stmt[] = {
 struct cfg_stmt auth_stmt[] = {
 	{ "port", CS_STMT, NULL, cfg_get_port, &auth_port, NULL, NULL },
 	{ "listen", CS_STMT, NULL, rad_cfg_listen_auth, NULL, NULL, NULL },
+	{ "forward", CS_STMT, NULL, rad_cfg_forward_auth, NULL, NULL, NULL },
 	{ "max-requests", CS_STMT, NULL,
 	  cfg_get_integer, &request_class[R_AUTH].max_requests, NULL, NULL },
 	{ "time-to-live", CS_STMT, NULL,
@@ -1255,12 +1323,17 @@ struct cfg_stmt auth_stmt[] = {
 	{ "checkrad-assume-logged", CS_STMT, NULL,
 	  cfg_get_boolean, &checkrad_assume_logged,
 	  NULL, NULL },
+	{ "reject-malformed-names", CS_STMT, NULL,
+	  cfg_get_boolean, &auth_reject_malformed_names,
+	  NULL, NULL },
 	{ "password-expire-warning", CS_STMT, NULL,
 	  cfg_get_integer, &warning_seconds,
 	  NULL, NULL },
 	{ "compare-attribute-flag", CS_STMT, NULL,
 	  cfg_get_integer, &auth_comp_flag,
 	  NULL, NULL },
+	{ "trace-rules", CS_STMT, NULL, cfg_get_boolean, &auth_trace_rules,
+	  NULL, NULL },
 	/* Obsolete statements */
 	{ "spawn", CS_STMT, NULL, cfg_obsolete, NULL, NULL, NULL },
 	{ NULL, }
@@ -1269,6 +1342,7 @@ struct cfg_stmt auth_stmt[] = {
 struct cfg_stmt acct_stmt[] = {
 	{ "port", CS_STMT, NULL, cfg_get_port, &acct_port, NULL, NULL },
 	{ "listen", CS_STMT, NULL, rad_cfg_listen_acct, NULL, NULL, NULL },
+	{ "forward", CS_STMT, NULL, rad_cfg_forward_auth, NULL, NULL, NULL },
 	{ "max-requests", CS_STMT, NULL,
 	  cfg_get_integer, &request_class[R_ACCT].max_requests,
 	  NULL, NULL },
@@ -1283,6 +1357,8 @@ struct cfg_stmt acct_stmt[] = {
 	{ "compare-attribute-flag", CS_STMT, NULL,
 	  cfg_get_integer, &acct_comp_flag,
 	  NULL, NULL },
+	{ "trace-rules", CS_STMT, NULL, cfg_get_boolean, &acct_trace_rules,
+	  NULL, NULL },
 	/* Obsolete statements */
 	{ "spawn", CS_STMT, NULL, cfg_obsolete, NULL, NULL, NULL },
 	{ NULL, }
@@ -1296,14 +1372,14 @@ struct cfg_stmt proxy_stmt[] = {
 };
 
 struct cfg_stmt config_syntax[] = {
-	{ "option", CS_BLOCK, NULL, NULL, NULL, option_stmt, NULL },
+	{ "option", CS_BLOCK, NULL, NULL, NULL, option_stmt, option_stmt_end },
 	{ "message", CS_BLOCK, NULL, NULL, NULL, message_stmt, NULL },
 	{ "logging", CS_BLOCK, logging_stmt_begin, logging_stmt_handler, NULL,
 	  logging_stmt, logging_stmt_end },
 	{ "auth", CS_BLOCK, auth_stmt_begin, NULL, NULL, auth_stmt, NULL },
 	{ "acct", CS_BLOCK, acct_stmt_begin, NULL, NULL, acct_stmt, NULL  },
 	{ "proxy", CS_BLOCK, NULL, NULL, NULL, proxy_stmt, NULL  },
-	{ "rewrite", CS_BLOCK, NULL, NULL, NULL, rewrite_stmt, NULL },
+	{ "rewrite", CS_BLOCK, rewrite_stmt_term, NULL, NULL, rewrite_stmt, NULL },
 	{ "filters", CS_BLOCK, filters_stmt_term, NULL, NULL, filters_stmt,
 	  NULL },
 #ifdef USE_DBM
