@@ -1,5 +1,7 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2002,2003 Sergey Poznyakoff
+   Copyright (C) 2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
   
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -66,7 +68,7 @@ struct request_class request_class[] = {
         { NULL, }
 };
 
-static LIST *request_list; /* List of REQUEST structures */
+static RAD_LIST *request_list; /* List of REQUEST structures */
 
 
 /* ************************* General-purpose functions ********************* */
