@@ -1,6 +1,8 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003 Sergey Poznyakoff
-  
+   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
+
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
@@ -44,15 +46,15 @@ static void snmpserv_before_config_hook(void *unused1, void *unused2);
 static void snmpserv_after_config_hook(void *arg, void *unused);
 static void snmp_tree_init();
 
-static LIST /* of ACL */ *snmp_acl;
-static LIST /* of Community */ *commlist;
+static RAD_LIST /* of ACL */ *snmp_acl;
+static RAD_LIST /* of Community */ *commlist;
 Server_stat *server_stat;
 struct radstat radstat;
 
 /* ************************************************************************ */
 /* Configuration file */
 
-static LIST /* of NETNAME */ *netlist;
+static RAD_LIST /* of NETNAME */ *netlist;
 
 static int
 _netname_cmp(const void *item, const void *data)
@@ -147,7 +149,7 @@ _netdef_cmp(const void *item, const void *data)
 	const NETDEF *nd = item;
 	const struct acl_closure *clos = data;
 
-	if (nd->ipaddr == (clos->ip & nd->netmask))
+	if (ip_addr_in_net_p(nd, clos->ip))
 		return 0;
 	return 1;
 }
@@ -181,7 +183,7 @@ check_acl(UINT4 ip, char *community)
 }
 
 void
-snmp_add_acl(Community *community, LIST /* of NETDEF */ *netlist)
+snmp_add_acl(Community *community, RAD_LIST /* of NETDEF */ *netlist)
 {
         ACL *acl;
 
@@ -218,7 +220,9 @@ snmp_stmt_begin(int finish, void *data, void *up_data)
 		snmp_free_communities();
 		snmp_free_acl();
 		_opened_snmp_sockets = 0;
-	} else if (radius_mode == MODE_DAEMON && !_opened_snmp_sockets) 
+	} else if (radius_mode == MODE_DAEMON
+		   && !_opened_snmp_sockets
+		   && snmp_port) 
 		udp_open(R_SNMP, INADDR_ANY, snmp_port, 1);
 	return 0;
 }
@@ -288,6 +292,12 @@ snmp_cfg_listen(int argc, cfg_value_t *argv,
 {
 	int i, errcnt = 0;
 	
+	if (argc == 2 && argv[1].type == CFG_BOOLEAN) {
+		if (argv[1].v.bool == 0)
+			snmp_port = 0;
+		return 0;
+	}
+	
 	for (i = 1; i < argc; i++) 
 		if (argv[i].type != CFG_HOST) {
 			cfg_type_error(CFG_HOST);
@@ -2141,8 +2151,9 @@ snmp_stat_nas(int num, enum mib_node_cmd cmd, struct nas_data *closure,
                         return -1;
                 }
 
+		/* FIXME: MIBS do not reflect netmask */
                 for (num = 0; num < 4; num++)
-                        closure->quad[num] = (nas->ipaddr >>
+                        closure->quad[num] = (nas->netdef.ipaddr >>
                                               (8*(3-num))) & 0xff;
                 
                 break;
@@ -2154,8 +2165,9 @@ snmp_stat_nas(int num, enum mib_node_cmd cmd, struct nas_data *closure,
                 if (num == 0) {
                         if (nas = nas_lookup_index(1))
                                 for (num = 0; num < 4; num++)
-                                        closure->quad[num] = (nas->ipaddr >>
-                                                              (8*(3-num))) & 0xff;
+                                        closure->quad[num] =
+						(nas->netdef.ipaddr >>
+                                                          (8*(3-num))) & 0xff;
                 }
                 break;
 
