@@ -1,5 +1,7 @@
 /* This file is part of GNU Radius.
-   Copyright (C) 2000,2001,2002,2003 Sergey Poznyakoff
+   Copyright (C) 2000,2001,2002,2003 Free Software Foundation, Inc.
+
+   Written by Sergey Poznyakoff
   
    GNU Radius is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,9 +38,11 @@
 # include <libguile.h>
 #endif
 
-static void sql_check_config(SQL_cfg *);
+static void sql_check_config(const char *, SQL_cfg *);
 static struct sql_connection *attach_sql_connection(int type);
 static void detach_sql_connection(int type);
+static void sql_conn_destroy(struct sql_connection **conn);
+static void sql_cache_destroy(struct sql_connection *conn);
 
 static char *getline();
 static int get_boolean(char *str, int *retval);
@@ -75,6 +79,9 @@ static struct sql_connection *sql_conn[SQL_NSERVICE];
 #define STMT_INTERFACE             22
 #define STMT_CHECK_ATTR_QUERY      23
 
+
+/* *********************** Configuration File Parser *********************** */
+
 static FILE  *sqlfd;
 static int line_no;
 static char *cur_line, *cur_ptr;
@@ -86,7 +93,7 @@ struct keyword sql_keyword[] = {
         { "port",               STMT_PORT },
         { "login",              STMT_LOGIN },
         { "password",           STMT_PASSWORD },
-        {"keepopen",            STMT_KEEPOPEN },
+        { "keepopen",           STMT_KEEPOPEN },
         { "idle_timeout",       STMT_IDLE_TIMEOUT },
         { "auth_max_connections", STMT_MAX_AUTH_CONNECTIONS },
         { "acct_max_connections", STMT_MAX_ACCT_CONNECTIONS },
@@ -219,10 +226,10 @@ sql_digest(SQL_cfg *cfg)
         *(int*)p = cfg->keepopen;
         p += sizeof(int);
         
-        *(int*)p = cfg->doauth;
+        *(int*)p = cfg->active[SQL_AUTH];
         p += sizeof(int);
         
-        *(int*)p = cfg->doacct;
+        *(int*)p = cfg->active[SQL_ACCT];
         p += sizeof(int);
         
         *(int*)p = cfg->port;
@@ -248,6 +255,17 @@ sql_digest_comp(char *d1, char *d2)
         return memcmp(d1, d2, len);
 }
 
+static int
+sql_cfg_empty_p(SQL_cfg *cfg)
+{
+	char *p;
+
+	for (p = (char*)cfg; p < (char*)(cfg + 1); p++)
+		if (*p)
+			return 0;
+	return 1;
+}
+
 int
 sql_cfg_comp(SQL_cfg *a, SQL_cfg *b)
 {
@@ -275,8 +293,8 @@ rad_sql_init()
         bzero(&new_cfg, sizeof(new_cfg));
         new_cfg.keepopen = 0;
         new_cfg.idle_timeout = 4*3600; /* four hours */
-        new_cfg.doacct = 0;
-        new_cfg.doauth = 0;
+        new_cfg.active[SQL_ACCT] = 0;
+        new_cfg.active[SQL_AUTH] = 0;
 
         sqlfile = mkfilename(radius_dir, "sqlserver");
         /* Open source file */
@@ -312,8 +330,8 @@ rad_sql_init()
                                        _("%s:%d: unknown host: %s"),
                                        sqlfile, line_no,
                                        cur_ptr);
-                                new_cfg.doacct = 0;
-                                new_cfg.doauth = 0;
+                                new_cfg.active[SQL_ACCT] = 0;
+                                new_cfg.active[SQL_AUTH] = 0;
                         } else {
                                 new_cfg.server = estrdup(cur_ptr);
                         }
@@ -326,8 +344,8 @@ rad_sql_init()
 				       "%s:%d: %s",
                                        sqlfile, line_no,
 				       _("number parse error"));
-                                new_cfg.doacct = 0;
-                                new_cfg.doauth = 0;
+                                new_cfg.active[SQL_ACCT] = 0;
+                                new_cfg.active[SQL_AUTH] = 0;
                         }
                         break;
                         
@@ -358,7 +376,7 @@ rad_sql_init()
                         break;
                         
                 case STMT_DOAUTH:       
-                        if (get_boolean(cur_ptr, &new_cfg.doauth))
+                        if (get_boolean(cur_ptr, &new_cfg.active[SQL_AUTH]))
                                 radlog(L_ERR,
                                        "%s:%d: %s",
                                        sqlfile, line_no,
@@ -366,7 +384,7 @@ rad_sql_init()
                         break;
                         
                 case STMT_DOACCT:
-                        if (get_boolean(cur_ptr, &new_cfg.doacct))
+                        if (get_boolean(cur_ptr, &new_cfg.active[SQL_ACCT]))
                                 radlog(L_ERR,
                                        "%s:%d: %s", 
                                        sqlfile, line_no,
@@ -453,11 +471,12 @@ rad_sql_init()
         obstack_free(&parse_stack, NULL);
 
         fclose(sqlfd);
+
+        sql_check_config(sqlfile, &new_cfg);
         efree(sqlfile);
 
-        sql_check_config(&new_cfg);
-
-        if (sql_cfg_comp(&new_cfg, &sql_cfg)) 
+        if (!sql_cfg_empty_p(&sql_cfg)
+	    && sql_cfg_comp(&new_cfg, &sql_cfg)) 
                 sql_flush();
 
         /* Free old configuration structure */
@@ -482,14 +501,11 @@ rad_sql_init()
         return 0;
 }
 
-static void
-sql_conn_destroy(struct sql_connection **conn)
+void
+radiusd_sql_clear_cache()
 {
-	if (*conn) {
-		disp_sql_disconnect(*conn);
-		efree(*conn);
-		*conn = NULL;
-	}
+	sql_cache_destroy(sql_conn[SQL_AUTH]);
+	sql_cache_destroy(sql_conn[SQL_ACCT]);
 }
 
 void
@@ -499,71 +515,94 @@ radiusd_sql_shutdown()
 	sql_conn_destroy(&sql_conn[SQL_ACCT]);
 }
 
-void
-sql_check_config(SQL_cfg *cfg)
+static void
+missing_statement(int prio, const char *filename, char *stmt)
 {
+	radlog(prio, _("%s: missing `%s' statement"),
+	       filename, stmt);
+}
+
+void
+sql_check_config(const char *filename, SQL_cfg *cfg)
+{
+	int doauth = cfg->active[SQL_AUTH];
+	int doacct = cfg->active[SQL_ACCT];
+
 #define FREE_IF_EMPTY(s) if (s && strcmp(s, "none") == 0) {\
                                 efree(s);\
                                 s = NULL;\
                          }
-        /*
-         * Check if we should do SQL authentication
-         */
-        if (cfg->doauth) {
-                FREE_IF_EMPTY(cfg->auth_query);
-                if (!cfg->auth_query) {
-                        radlog(L_ERR,
-                            _("disabling SQL auth: no auth_query specified"));
-                        cfg->doauth = 0;
-                }
-                if (!cfg->group_query) {
-                        radlog(L_WARN,
-                               _("SQL auth: no group_query specified"));
-                }
+
+	/* Check general SQL setup */
+	if (cfg->active[SQL_AUTH] || cfg->active[SQL_ACCT]) {
+		if (!cfg->server) {
+			missing_statement(L_ERR, filename, "server");
+			doacct = doauth = 0;
+		}
+
+		if (!cfg->login) {
+			missing_statement(L_ERR, filename, "login");
+			doacct = doauth = 0;
+		}
+
+		if (!cfg->password) {
+			missing_statement(L_ERR, filename, "password");
+			doacct = doauth = 0;
+		}
+	}
+	
+	/* Check SQL authentication setup */
+        if (cfg->active[SQL_AUTH]) {
+		if (!cfg->auth_db) {
+			missing_statement(L_ERR, filename, "auth_db");
+			doauth = 0;
+		}
+
+		FREE_IF_EMPTY(cfg->auth_query);
+		if (!cfg->auth_query) {
+			missing_statement(L_ERR, filename, "auth_query");
+			doauth = 0;
+		}
+			
+                if (!cfg->group_query) 
+			missing_statement(L_WARN, filename, "group_query");
                 FREE_IF_EMPTY(cfg->group_query);
         }
-        /*
-         * Check if we should do SQL accounting
-         */
-        if (cfg->doacct) {
-                if (!cfg->acct_start_query) {
-                        radlog(L_WARN,
-                               _("SQL acct: no acct_start_query specified"));
-                }
+
+	/* Check SQL accounting setup */
+        if (cfg->active[SQL_ACCT]) {
+		if (!cfg->acct_db) {
+			missing_statement(L_ERR, filename, "acct_db");
+			doacct = 0;
+		}
+                if (!cfg->acct_start_query) 
+			missing_statement(L_WARN, filename,
+					  "acct_start_query");
                 FREE_IF_EMPTY(cfg->acct_start_query);
 
-                if (!cfg->acct_stop_query) {
-                        radlog(L_ERR,
-                     _("disabling SQL acct: no acct_stop_query specified"));
-                        cfg->doacct = 0;
-                }
-                FREE_IF_EMPTY(cfg->acct_stop_query);
+		if (!cfg->acct_stop_query) {
+			missing_statement(L_ERR, filename, "acct_stop_query");
+			doacct = 0;
+		}
+		FREE_IF_EMPTY(cfg->acct_stop_query);
 
-                if (!cfg->acct_nasdown_query) {
-                        radlog(L_WARN,
-                     _("SQL acct: no acct_nasdown_query specified"));
-                }
+                if (!cfg->acct_nasdown_query) 
+			missing_statement(L_WARN, filename,
+					  "acct_nasdown_query");
                 FREE_IF_EMPTY(cfg->acct_nasdown_query);
 
-                if (!cfg->acct_nasup_query) {
-                        radlog(L_WARN,
-                     _("SQL acct: no acct_nasup_query specified"));
-                }
+                if (!cfg->acct_nasup_query)
+			missing_statement(L_WARN, filename,
+					  "acct_nasup_query");
                 FREE_IF_EMPTY(cfg->acct_nasup_query);
         }
-        
-        debug(1, ("SQL init using: %s:%d,%s,%s,%s,%d,%ld,%d,%d",
-               cfg->server,
-               cfg->port,
-               cfg->login,
-               cfg->acct_db,
-               cfg->auth_db,
-               cfg->keepopen,
-               cfg->idle_timeout,
-               cfg->doacct,
-               cfg->doauth));
-}
 
+	/* Inform the user */
+	if (cfg->active[SQL_AUTH] != doauth)
+		radlog(L_NOTICE, _("disabling SQL authentication"));
+	if (cfg->active[SQL_ACCT] != doacct)
+		radlog(L_NOTICE, _("disabling SQL accounting"));
+}
 
 void
 sql_flush()
@@ -574,6 +613,131 @@ sql_flush()
 	radiusd_sql_shutdown();
 }
 
+
+/* ********************************* SQL Cache ***************************** */
+static void
+sql_result_destroy(SQL_RESULT *res)
+{
+	size_t i;
+	
+	if (!res)
+		return;
+	efree(res->query);
+	for (i = 0; i < res->ntuples; i++) {
+		size_t j;
+		for (j = 0; j < res->nfields; j++) 
+			efree(res->tuple[i][j]);
+		efree(res->tuple[i]);
+	}
+	efree(res->tuple);
+	efree(res);
+}
+
+static size_t
+sql_cache_level(struct sql_connection *conn)
+{
+	if (conn->tail < conn->head)
+		return conn->tail + SQL_CACHE_SIZE - conn->head;
+	else
+		return conn->tail - conn->head;
+}
+
+static void
+sql_cache_destroy(struct sql_connection *conn)
+{
+	size_t i;
+
+	if (!conn)
+		return;
+	for (; sql_cache_level(conn) > 0;
+	     conn->head = (conn->head + 1) % SQL_CACHE_SIZE)
+		sql_result_destroy(conn->cache[conn->head]);
+}
+
+static void
+sql_cache_insert(struct sql_connection *conn, SQL_RESULT *res)
+{
+	debug(20,
+	      ("cache: %d,(%d,%d)",
+	       sql_cache_level(conn),conn->head,conn->tail));
+	if (sql_cache_level(conn) >= SQL_CACHE_SIZE-1) {
+		sql_result_destroy(conn->cache[conn->head]);
+		conn->head = (conn->head + 1) % SQL_CACHE_SIZE;
+	        debug(20,("head: %d,%d", conn->head,conn->tail));
+	}
+	debug(20,("inserting at pos %d", conn->tail));
+	conn->cache[conn->tail] = res;
+	conn->tail = (conn->tail + 1) % SQL_CACHE_SIZE;
+	debug(20,("tail: %d,%d", conn->head,conn->tail));
+}
+
+static SQL_RESULT *
+sql_cache_retrieve(struct sql_connection *conn, char *query)
+{
+        void *data;
+	SQL_RESULT *res;
+	size_t i;
+	
+	debug(20,("query: %s",query));
+	res = emalloc(sizeof(*res));
+	res->query = estrdup(query);
+	res->nfields = res->ntuples = 0;
+	res->tuple = NULL;
+        data = disp_sql_exec(conn, query);
+        if (!data) {
+		sql_cache_insert(conn, res);
+                return res;
+	}
+	
+	if (disp_sql_num_tuples(conn, data, &res->ntuples)
+	    || disp_sql_num_columns(conn, data, &res->nfields)) {
+		disp_sql_free(conn, data);
+		res->nfields = res->ntuples = 0;
+		sql_cache_insert(conn, res);
+		return res;
+	}
+
+	res->tuple = emalloc(sizeof(res->tuple[0]) * res->ntuples);
+        for (i = 0; i < res->ntuples
+		     && disp_sql_next_tuple(conn, data) == 0; i++) {
+		int j;
+
+		res->tuple[i] = emalloc(sizeof(res->tuple[0][0]) * res->nfields);
+			
+		for (j = 0; j < res->nfields; j++) {
+			char *p = disp_sql_column(conn, data, j);
+			chop(p);
+			res->tuple[i][j] = estrdup(p);
+		}
+	}
+		
+        disp_sql_free(conn, data);
+
+	sql_cache_insert(conn, res);
+	
+        return res;
+}
+
+static SQL_RESULT *
+sql_cache_lookup(struct sql_connection *conn, char *query)
+{
+	size_t i;
+
+	debug(20,("looking up %s", query));
+	for (i = conn->head; i != conn->tail;
+	     i = (i + 1) % SQL_CACHE_SIZE) {
+		if (strcmp(conn->cache[i]->query, query) == 0) {
+			debug(20,("found at %d", i));
+			return conn->cache[i];
+		}
+	}
+	debug(20,("NOT FOUND"));
+	return NULL;
+}
+
+
+/* *********************** SQL Connection Handling ************************* */
+
 struct sql_connection *
 attach_sql_connection(int type)
 {
@@ -590,6 +754,8 @@ attach_sql_connection(int type)
                 conn->last_used = now;
                 conn->type = type;
 
+		conn->head = 0;
+		conn->tail = 0;
                 sql_conn[type] = conn;
         }
 
@@ -622,6 +788,20 @@ detach_sql_connection(int type)
         }
 }
 
+static void
+sql_conn_destroy(struct sql_connection **conn)
+{
+	if (*conn) {
+		disp_sql_disconnect(*conn);
+		sql_cache_destroy(*conn);
+		efree(*conn);
+		*conn = NULL;
+	}
+}
+
+
+/* ************************* Interface Functions *************************** */
+
 /*ARGSUSED*/
 void
 rad_sql_cleanup(int type, void *req ARG_UNUSED)
@@ -645,14 +825,14 @@ rad_sql_acct(RADIUS_REQ *radreq)
         char *query_name;
 	int log_facility = 0;
 	
-        if (!sql_cfg.doacct)
+        if (!sql_cfg.active[SQL_ACCT])
                 return;
 
         if ((pair = avl_find(radreq->request, DA_ACCT_STATUS_TYPE)) == NULL) {
                 /* should never happen!! */
                 radlog_req(L_ERR, radreq,
                            _("no Acct-Status-Type attribute in rad_sql_acct()"));
-                return ;
+                return;
         }
         status = pair->avp_lvalue;
 
@@ -747,7 +927,7 @@ rad_sql_pass(RADIUS_REQ *req, char *authdata)
         char *query;
         struct obstack stack;
         
-        if (sql_cfg.doauth == 0) {
+        if (sql_cfg.active[SQL_AUTH] == 0) {
                 radlog(L_ERR,
                        _("SQL Auth specified in users file, but not in sqlserver file"));
                 return NULL;
@@ -780,12 +960,12 @@ rad_sql_checkgroup(req, groupname)
 {
         int   rc = -1;
         struct sql_connection *conn;
-        void *data;
-        char *p;
         char *query;
         struct obstack stack;
+	SQL_RESULT *res;
+	size_t i;
 	
-        if (sql_cfg.doauth == 0 || sql_cfg.group_query == NULL) 
+        if (sql_cfg.active[SQL_AUTH] == 0 || sql_cfg.group_query == NULL) 
                 return -1;
 
         conn = attach_sql_connection(SQL_AUTH);
@@ -795,21 +975,22 @@ rad_sql_checkgroup(req, groupname)
         obstack_init(&stack);
 	
         query = radius_xlate(&stack, sql_cfg.group_query, req, NULL);
+	res = sql_cache_lookup(conn, query);
+	if (!res) {
+		res = sql_cache_retrieve(conn, query);
+		if (!res)
+			return rc;
+	}
+        obstack_free(&stack, NULL);
 
-        data = disp_sql_exec(conn, query);
-	if (data) {
-		while (rc != 0
-		       && disp_sql_next_tuple(conn, data) == 0) {
-			if ((p = disp_sql_column(conn, data, 0)) == NULL)
-				break;
-			chop(p);
-			if (strcmp(p, groupname) == 0)
-				rc = 0;
-		}
-		disp_sql_free(conn, data);
+	if (res->ntuples == 0 || res->nfields == 0)
+		return rc;
+
+	for (i = 0; rc != 0 && i < res->ntuples; i++) {
+		if (strcmp(res->tuple[i][0], groupname) == 0)
+			rc = 0;
 	}
 	
-        obstack_free(&stack, NULL);
         return rc;
 }
 
@@ -819,52 +1000,53 @@ rad_sql_retrieve_pairs(struct sql_connection *conn,
 		       VALUE_PAIR **return_pairs,
 		       int op_too)
 {
-        void *data;
-        int i;
-        
-        data = disp_sql_exec(conn, query);
-        if (!data) 
-                return 0;
-        
-        for (i = 0; disp_sql_next_tuple(conn, data) == 0; i++) {
-                VALUE_PAIR *pair;
-                char *attribute;
-                char *value;
-                int op;
-                
-                if (!(attribute = disp_sql_column(conn, data, 0))
-                    || !(value =  disp_sql_column(conn, data, 1)))
-                        break;
-                if (op_too) {
-                        char *opstr;
-                        opstr = disp_sql_column(conn, data, 2);
-                        if (!opstr)
-                                break;
-                        chop(opstr);
+	SQL_RESULT *res;
+	size_t i;
+	LOCUS loc;
+	
+	res = sql_cache_lookup(conn, query);
+	if (!res) {
+		res = sql_cache_retrieve(conn, query);
+		if (!res)
+			return 0;
+	}
+
+	if (res->ntuples == 0
+	    || res->nfields < 2
+	    || (op_too && res->nfields < 3))
+		return 0;
+	
+        for (i = 0; i < res->ntuples; i++) {
+		int op;
+		VALUE_PAIR *pair;
+		
+		if (op_too) {
+			char *opstr = res->tuple[i][2];
                         op = str_to_op(opstr);
                         if (op == NUM_OPERATORS) {
                                 radlog(L_NOTICE,
                                        _("SQL: invalid operator: %s"), opstr);
                                 continue;
                         }
-                } else
-                        op = OPERATOR_EQUAL;
+		} else
+			op = OPERATOR_EQUAL;
 
-                chop(attribute);
-                chop(value);
-                
-                pair = install_pair(__FILE__, __LINE__, attribute, op, value);
+		loc.file = __FILE__;
+		loc.line = __LINE__;
+                pair = install_pair(&loc,
+				    res->tuple[i][0],
+				    op,
+				    res->tuple[i][1]);
                 
                 if (pair) {
                         avl_merge(return_pairs, &pair);
                         avl_free(pair);
                 }
-        }
+	}		
 
-        disp_sql_free(conn, data);
-        return i;
+	return i;
 }
-
+			
 int
 rad_sql_reply_attr_query(RADIUS_REQ *req, VALUE_PAIR **reply_pairs)
 {
@@ -873,7 +1055,7 @@ rad_sql_reply_attr_query(RADIUS_REQ *req, VALUE_PAIR **reply_pairs)
         int rc;
         struct obstack stack;
         
-        if (sql_cfg.doauth == 0 || !sql_cfg.reply_attr_query)
+        if (sql_cfg.active[SQL_AUTH] == 0 || !sql_cfg.reply_attr_query)
                 return 0;
         
         conn = attach_sql_connection(SQL_AUTH);
@@ -894,7 +1076,7 @@ rad_sql_check_attr_query(RADIUS_REQ *req, VALUE_PAIR **return_pairs)
         int rc;
         struct obstack stack;
         
-        if (sql_cfg.doauth == 0 || !sql_cfg.check_attr_query)
+        if (sql_cfg.active[SQL_AUTH] == 0 || !sql_cfg.check_attr_query)
                 return 0;
         
         conn = attach_sql_connection(SQL_AUTH);
